#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy
import time
import itertools
import traceback

from traits.api import HasTraits, Int, Float, \
     Bool, Property, Array, Event, List, cached_property, Str,\
     Instance, on_trait_change, Trait, Enum, Title, Complex, Dict

from traitsui.api import View, Item, Tabbed, VGroup, Include, \
    Group

from tvtk.api import tvtk

from raypier.core.ctracer import RayCollection, GaussletCollection, Ray, ray_dtype, GAUSSLET_, PARABASAL_
from raypier.utils import normaliseVector, Range, TupleVector, Tuple, \
            UnitTupleVector, UnitVectorTrait
from raypier.bases import RaypierObject, NumEditor, BaseRayCollection

Vector = Array(shape=(3,))


class BaseBase(HasTraits, RaypierObject):
    pass


class BaseRaySource(BaseBase):
    abstract=True
    subclasses = set()
    name = Title("Ray Source", allow_selection=False)
    
    ### Updated to the current timestamp whenever the traced rays are changed (i.e. after a tracing operation)
    _mtime = Float(0.0)
    
    ### The update event trigger a new tracing operation
    update = Event()
    display = Enum("pipes", "wires", "hidden")
    
    ### The render event signals that the VTK view needs to be re-rendered
    render = Event()
    mapper = Instance(tvtk.PolyDataMapper, (), transient=True)
    
    wavelength_list = List() #List of wavelengths (floats) given in microns
    
    input_rays = Property(Instance(BaseRayCollection), depends_on="max_ray_len")
    traced_rays = List(BaseRayCollection, transient=True)
    
    InputDetailRays = Property(Instance(BaseRayCollection), depends_on="input_rays")
    TracedDetailRays = List(BaseRayCollection, transient=True)

    detail_resolution = Int(32)

    max_ray_len = Float(200.0)
    
    scale_factor = Float(0.1)
    
    show_start = Bool(True)
    show_polarisation = Bool(False)
    show_direction = Bool(False)
    show_normals = Bool(False)
    opacity = Range(0.0,1.0,1.0)
    
    ### A dict which maps each RayCollection instance in the traced_rays
    ### list to a list/array of bools of the same length as the RayCollection
    ### This bool array indicates if a ray should be hidden in the display
    ray_mask = Dict()
    
    #idx selector for the input ways which should be visualised
    #making this transient because pyYAML fails to serialise arrays
    view_ray_ids = Trait(None, Array(dtype=numpy.int), transient=True)
    
    tube = Instance(tvtk.TubeFilter, (), transient=True)
    sphere = Instance(tvtk.SphereSource, (), transient=True)
    data_source = Instance(tvtk.ProgrammableSource, transient=True)
    
    normals_source = Instance(tvtk.ProgrammableSource, transient=True)
    normals_actor = Instance(tvtk.Actor, transient=True)
    normals_glyph = Instance(tvtk.Glyph3D, transient=True)
    
    ray_actor = Instance(tvtk.Actor, transient=True)
    start_actor = Instance(tvtk.Actor, transient=True)
    #field_actor = Instance(tvtk.Actor)
    
    export_pipes = Bool(False, desc="if set, rays are exported to STEP files as pipes")
    
    actors = Property(List, transient=True)
    
    vtkproperty = Instance(tvtk.Property, (), {'color':(1,0.5,0), 'opacity': 1.0}, transient=True)
    
    display_grp = VGroup(Item('display'),
                       Item('show_start'),
                       Item('show_normals'),
                       Item('scale_factor'),
                       Item('export_pipes',label="export as pipes"),
                       Item('opacity', editor=NumEditor),
                       label="Display")
    
    traits_view = View(Item('name', show_label=False),
                       Tabbed(Include('geom_grp'),
                              display_grp)
                       )
    
    def __repr__(self):
        return self.name
    
    def get_sequence_to_face(self, face):
        """returns a list of list of Face objects, those encountered
        on the route to the target face"""
        #find the first RayCollection which contains the target face
        for rays in self.traced_rays[1:]:
            faces = list(rays.face)
            if face in faces:
                break
        else:
            raise ValueError("no rays trace to this face")
        
        seq = []
        
        #now iterate back up the ray-tree collecting only the faces
        #on the path to the target face
        parent = rays.parent
        ids = rays.parent_ids[face==rays.face]
        while True:
            if parent == self.traced_rays[0]:
                break
            faces = list(numpy.unique1d(parent.face[ids]))
            seq.append(faces)
            ids = parent.parent_ids[ids]
            rays = parent
            parent = rays.parent
        seq.reverse()
        return seq


    def get_ray_list_by_id(self):
        """return a list of lists of dictionaries where the index of the outer list is the ray id (defined by order encountered)
        , and the inner index is the recursion # backwards (0 being the ray that dies, 1 is its parent, and so on) 
        and the dictionary keys are the attributes of the ray object"""
        result = []
        #keys is copy and pasted from the dtype of a ray object
        keys = ['origin','direction','normal','E_vector','refractive_index','E1_amp','E2_amp','length','wavelength','parent_idx','end_face_idx']
        #start at the last ray collection and work backwards
        temp = list(self.traced_rays)    #otherwise its a TraitListObject
        raycollections = []             #make data a list of lists of lists
        for raycollection in temp:
            tmp = []
            for x in raycollection.copy_as_array():
                tmp.append(list(x))
            raycollections.append(tmp)

        length = len(raycollections)-1

        for i,collection in enumerate(reversed(raycollections)):
            for ray in collection:
                if ray:                  #rays already accounted for are set to none
                    lst = []             #list to contain all relevant dictionaries for each ray
                    r = {}               #the first dictionary
                    for x,att in enumerate(ray):
                        r[keys[x]] = att 
                    lst.append(r)
                    cntr = length-i
                    cntr -= 1
                    parent_idx = r['parent_idx']
                    while cntr >= 0:     #iterate back through the ray collections
                        r = {}          #and repeat what we did for the first ray
                        arr = raycollections[cntr][parent_idx]
                        for x,att in enumerate(arr):
                            r[keys[x]] = att 
                        lst.append(r.copy())
                        raycollections[cntr][parent_idx] = None     #mark this ray done
                        cntr -= 1
                        parent_idx = r['parent_idx']
                    result.append(lst)

        
        return result

    def eval_angular_spread(self, idx):
        """A helper method to evaluate the angular spread of a ray-segment.
        
        :param idx: the index of the RayCollection in the TracedRay list 
                    to analyse
                    
        :returns: the average angle from the mean ray, in degrees
        
        """
        rays = self.traced_rays[idx]
        ave_dir = normaliseVector(rays.direction.mean(axis=0))
        dotprod = (ave_dir[numpy.newaxis,:] * rays.direction).sum(axis=1)
        angles = numpy.arccos(dotpod)*180 / numpy.pi
        return angles.mean()
    
    def make_step_shape(self):
        if self.export_pipes:
            from raypier.step_export import make_rays_both as make_rays
        else:
            from raypier.step_export import make_rays_wires as make_rays
        return make_rays(self.traced_rays, self.scale_factor), "red"
    
    def _traced_rays_changed(self):
        self.data_source.modified()
        self.normals_source.modified()
        self._mtime = time.monotonic()
        print("SET MTIME")
        
    def _get_input_rays(self):
        return None
    
    def _input_rays_changed(self):
        self.update=True
        
    def _scale_factor_changed(self, scale):
        self.tube.radius = scale
        self.normals_glyph.scale_factor=self.scale_factor*10
        self.sphere.radius = self.scale_factor*3
        self.render = True
        
    def _show_normals_changed(self):
        self.normals_source.modified()
        self.render=True
        
    def _ray_mask_changed(self):
        self.update = True
        
    def _opacity_changed(self):
        self.vtkproperty.opacity = self.opacity
        self.render = True
    
    def _get_actors(self):
        actors = [self.ray_actor, self.start_actor, self.normals_actor]
        return actors
        #return actors[:2]  #makeshift turning off of normal glyphs
        
    def _normals_glyph_default(self):
        glyph =  tvtk.Glyph3D(scale_factor=self.scale_factor*10,
                                    vector_mode='use_normal'
                                    )
        return glyph
    
    def _normals_source_default(self):
        source = tvtk.ProgrammableSource()
        def execute():
            if not self.show_normals:
                return
            output = source.poly_data_output
            points = numpy.vstack([p.origin for p in self.traced_rays[1:]])
            normals = numpy.vstack([p.normal for p in self.traced_rays[1:]])
            output.points = points
            output.point_data.normals = normals
            #print("calc normals GLYPH")
        source.set_execute_method(execute)
        return source
        
    def _normals_actor_default(self):
        source = self.normals_source
        glyph = tvtk.ArrowSource(tip_radius=0.3, shaft_radius=0.1)
        glyph_filter = self.normals_glyph
        glyph_filter.set_input_connection(source.output_port)
        glyph_filter.set_source_connection(glyph.output_port)
        map = tvtk.PolyDataMapper(input_connection=glyph_filter.output_port)
        act = tvtk.Actor(mapper=map)
        act.property.color = (0.0,1.0,0.0)
        return act
    
    def _data_source_default(self):
        source = tvtk.ProgrammableSource()
        def execute():
            output = source.poly_data_output
            pointArrayList = []
            mask = self.ray_mask
            gausslet = isinstance(self.input_rays, GaussletCollection)
            for rays in self.traced_rays:
                vis = mask.get(rays, [])
                if len(vis) != len(rays):
                    vis = itertools.repeat(True)
                if gausslet:
                    start_pos = [r.base_ray.origin for r,v in zip(rays, vis) if v]
                    end_pos = [r.base_ray.termination for r,v in zip(rays, vis) if v]
                else:
                    start_pos = [r.origin for r,v in zip(rays, vis) if v]
                    end_pos = [r.termination for r,v in zip(rays, vis) if v]
                #print "start", start_pos
                #print "end", end_pos
                interleaved = numpy.array([start_pos, end_pos]).swapaxes(0,1).reshape(-1,3)
                pointArrayList.append(interleaved)
            if pointArrayList:
                points = numpy.vstack(pointArrayList)
                output.points=points
                lines = list([i,i+1] for i in range(0,points.shape[0],2))
                output.lines = lines
        source.set_execute_method(execute)
        return source
    
    def _ray_actor_default(self):
        tube = self.tube
        tube.radius = self.scale_factor
        tube.input_connection=self.data_source.output_port
        tube.number_of_sides = 20
        
        map = self.mapper
        act = tvtk.Actor(mapper=map)
        act.visibility = True
        display = self.display
        if display=="pipes":
            map.input_connection=tube.output_port
        elif display=="wires":
            map.input_connection = self.data_source.output_port
        else:
            act.visibility = False
        prop = self.vtkproperty
        prop.opacity = self.opacity
        if prop:
            act.property = prop
        return act
    
    def _display_changed(self, vnew):
        actors = self.actors
        for act in actors:
            act.visibility = True
        if vnew=="pipes":
            self.mapper.input_connection = self.tube.output_port
        elif vnew=="wires":
            self.mapper.input_connection = self.data_source.output_port
        elif vnew=="hidden":
            for act in actors:
                act.visibility = False
        self.render = True
    
    def _start_actor_default(self):
        map = tvtk.PolyDataMapper(input_connection=self.sphere.output_port)
        act = tvtk.Actor(mapper=map)
        return act
    
    
class SingleRaySource(BaseRaySource):
    abstract = False
    origin = Tuple((0.,0.,0.))
    direction = UnitTupleVector
    
    E_vector = UnitVectorTrait((1.,0.,0.), editor_traits={'cols':3,
                                'labels':['x','y','z']})
    E1_amp = Complex(1.0+0.0j)
    E2_amp = Complex(0.0+0.0j)
    wavelength = Float(0.78) # in microns
    
    view_ray_ids = numpy.arange(1)
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, max_ray_len, E_vector, E1_amp, E2_amp")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       label="Geometry")
    
    def _wavelength_list_default(self):
        return [self.wavelength,]
    
    @on_trait_change("wavelength")
    def _do_wavelength_changed(self):
        self.wavelength_list = [self.wavelength]
    
    @on_trait_change("origin, direction, max_ray_len")
    def on_update(self):
        self.data_source.modified()
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        origin = numpy.array(self.origin)
        direction = numpy.array(self.direction)
        max_axis = numpy.abs(direction).argmax()
        if max_axis==0:
            v = numpy.array([0.,1.,0.])
        else:
            v = numpy.array([1.,0.,0.])
        d1 = numpy.cross(direction, v)
        d1 = normaliseVector(d1)
        d2 = numpy.cross(direction, d1)
        d2 = normaliseVector(d2)

        E_vector = numpy.cross(direction, self.E_vector)
        E_vector = numpy.cross(E_vector, direction)

        rays = RayCollection(1)
        ray = Ray()
        ray.origin = origin
        ray.direction = direction
        ray.wavelength_idx = 0
        ray.E_vector = E_vector
        ray.E1_amp = self.E1_amp
        ray.E2_amp = self.E2_amp
        ray.refractive_index = 1.0
        ray.normals = (0.,1.,0.) #What is this for?
        ray.parent_idx = -1
        ray.end_face_idx = -1
        
        rays.add_ray(ray)
        return rays
    
    
class BroadbandRaySource(SingleRaySource):
    """Creates a number of coincident rays covering a range of wavelengths.
    The wavelength spacings may be uniform in wavelength, or uniform in frequency.
    Wavelength is given in microns and is the in-vacuo wavelength.
    """
    number = Int(200, auto_set=False, enter_set=True)
    wavelength_start = Float(1.50, editor=NumEditor)
    wavelength_end = Float(1.60, editor=NumEditor)
    uniform_deltaf = Bool(True)
    
    view_ray_ids = numpy.arange(200)
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, number, wavelength_start, "
                         "wavelength_end, uniform_deltaf, max_ray_len, E_vector")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item('number'),
                       Item('wavelength_start'),
                       Item('wavelength_end'),
                       Item("uniform_deltaf"),
                       label="Geometry")
    
    @on_trait_change("direction, number, wavelength_start, wavelength_end, uniform_deltaf, max_ray_len")
    def on_update(self):
        self.data_source.modified()
        self.update=True
        
    @on_trait_change("number, wavelength_start, wavelength_end, uniform_deltaf")
    def _do_wavelength_changed(self):
        if self.uniform_deltaf:
            f_start = 1./self.wavelength_start
            f_end = 1./self.wavelength_end
            wavelengths = 1./numpy.linspace(f_start, f_end, self.number)
        else:
            wavelengths = numpy.linspace(self.wavelength_start, self.wavelength_end,
                                         self.number)
        self.wavelength_list = list(wavelengths)
        
    @cached_property
    def _get_input_rays(self):
        origin = numpy.array(self.origin)
        direction = numpy.array(self.direction)
        count = self.number

        E_vector = numpy.cross(self.E_vector, direction)
        E_vector = numpy.cross(E_vector, direction)
        
        wavelen_idx = numpy.arange(count)

        ray_data = numpy.zeros(count, dtype=ray_dtype)            
        ray_data['origin'] = origin.reshape(-1,3)
        ray_data['direction'] = direction.reshape(-1,3)
        ray_data['wavelength_idx'] = wavelen_idx
        ray_data['E_vector'] = [normaliseVector(E_vector)]
        ray_data['E1_amp'] = self.E1_amp
        ray_data['E2_amp'] = self.E2_amp
        ray_data['refractive_index'] = 1.0+0.0j
        ray_data['normal'] = [[0,1,0]]
        rays = RayCollection.from_array(ray_data)
        return rays
    
    
class ParallelRaySource(SingleRaySource):
    
    number = Int(20, auto_set=False, enter_set=True)
    radius = Float(10.,editor=NumEditor)
    rings = Range(0,50,3, editor_traits={'mode':'spinner'})
    
    view_ray_ids = numpy.arange(20)
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, number, rings, radius, max_ray_len, E_vector")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item('number'),
                       Item('rings'),
                       Item('radius'),
                       label="Geometry")
    
    
    @on_trait_change("direction, number, rings, radius, max_ray_len")
    def on_update(self):
        self.data_source.modified()
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        origin = numpy.array(self.origin)
        direction = numpy.array(self.direction)
        count = self.number
        rings = self.rings
        radius = self.radius
        max_axis = numpy.abs(direction).argmax()
        if max_axis==0:
            v = numpy.array([0.,1.,0.])
        else:
            v = numpy.array([1.,0.,0.])
        d1 = numpy.cross(direction, v)
        d1 = normaliseVector(d1)
        d2 = numpy.cross(direction, d1)
        d2 = normaliseVector(d2)

        E_vector = numpy.cross(self.E_vector, direction)
        E_vector = numpy.cross(E_vector, direction)

        ray_data = numpy.zeros((rings*count)+1, dtype=ray_dtype)
        if rings:
            radii = ((numpy.arange(rings)+1)*(radius/rings))[:,None,None]
            angles = (numpy.arange(count)*(2*numpy.pi/count))[None,:,None]
            offsets = radii*(d1*numpy.sin(angles) + d2*numpy.cos(angles)) 
            offsets.shape = (-1,3)
            
            ray_data['origin'][1:] = offsets
        ray_data['origin'] += origin
        ray_data['direction'] = direction
        ray_data['wavelength_idx'] = 0
        ray_data['E_vector'] = [normaliseVector(E_vector)]
        ray_data['E1_amp'] = self.E1_amp
        ray_data['E2_amp'] = self.E2_amp
        ray_data['refractive_index'] = 1.0+0.0j
        ray_data['normal'] = [[0,1,0]]
        rays = RayCollection.from_array(ray_data)
        return rays
    
    
class RectRaySource(BaseRaySource):
    """ rays from a rectangular aperture """ 
    origin = Tuple((0.,0.,0.))
    abstract = False
    direction = UnitTupleVector
    number = Int(20,auto_set=False,enter_set=True)		#total rays is n^2
    length = Float(10.,editor=NumEditor)
    width = Float(10,editor=NumEditor)
    rings = Range(1,50,3, editor_traits={'mode':'spinner'})
    theta = Float(0.,editor=NumEditor)
    randomness = Bool(False)
    
    view_ray_ids = numpy.arange(20)
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, number, length, width, theta, max_ray_len")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item('number'),
                       Item('length'),
                       Item('width'),
                       Item('theta'),
		       Item('randomness'),
                       label="Geometry")
    
    
    @on_trait_change("focus, direction, number, length, width, theta, max_ray_len")
    def on_update(self):
        self.data_source.modified()
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        from .utils import z_rotation
        origin = numpy.array(self.origin)
        direction = numpy.array(self.direction)
        count = self.number
        length = self.length
        width = self.width
        randomness = self.randomness
        max_axis = numpy.abs(direction).argmax()
        if max_axis==0:
            v = numpy.array([0.,1.,0.])
        else:
            v = numpy.array([1.,0.,0.])
        d1 = numpy.cross(direction, v)
        d1 = normaliseVector(d1)
        d2 = numpy.cross(direction, d1)
        d2 = normaliseVector(d2)
        a = length/2
        b = width/2
        X_range = numpy.linspace(-a, a, count)
        Y_range = numpy.linspace(-b, b, count)

        ray_data = numpy.zeros(count**2, dtype=ray_dtype)
        offsets = numpy.zeros([X_range.size*Y_range.size,3])
        
        if randomness:
            from random import uniform
        
        for i,x in enumerate(X_range):
            for j,y in enumerate(Y_range):
                if randomness:
                    x = x + uniform(-a/count, a/count)
                    y = y + uniform(-b/count, b/count)
                point = d1 * x + d2 * y
                block = i * Y_range.size
                offsets[block + j] = point
        
        origins = numpy.array([origin + offset for offset in offsets])
        
        dirmatrix = numpy.matrix(direction)
        #print "dirmatrix: ",z_rotation(numpy.radians(self.theta))
        raydir = z_rotation(numpy.radians(self.theta))*(dirmatrix.T)
        directions = numpy.ones_like(origins) * numpy.array(raydir.T)

        ray_data['origin'] = origins
        ray_data['direction'] = directions
        ray_data['E_vector'] = [[1,0,0]]
        ray_data['E1_amp'] = 1.0 + 0.0j
        ray_data['E2_amp'] = 0.0
        ray_data['refractive_index'] = 1.0+0.0j
        ray_data['normal'] = [[0,1,0]]
        rays = RayCollection.from_array(ray_data)
        return rays
    
    
class GaussianBeamRaySource(SingleRaySource):
    """
    Launches a set of skew rays simulating the propagation of a Gaussian beam.
    See Paul D Colbourne - Proc. SPIE 9293, International Optical Design Conference 2014, 92931S
    and refs theirin.
    """
    beam_waist = Float(100.0) #in microns
    number = Range(2,64,16)
    working_distance = Float(0.0) #the distance from the beam waist to the source location
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, number, beam_waist, wavelength, "
                         "max_ray_len, E_vector, working_distance")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item('number'),
                       Item("wavelength"),
                       Item('beam_waist', tooltip="1/e^2 beam intensity radius, in microns"),
                       Item('working_distance', tooltip="The distance from the source to the beam waist"),
                       label="Geometry")
    
    
    @on_trait_change("direction, number, beam_waist, max_ray_len, working_distance")
    def on_update(self):
        self.data_source.modified()
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        try:
            origin = numpy.array(self.origin)
            direction = numpy.array(self.direction)
            count = self.number
            radius = self.beam_waist/2000.0 #convert to mm
            max_axis = numpy.abs(direction).argmax()
            if max_axis==0:
                v = numpy.array([0.,1.,0.])
            else:
                v = numpy.array([1.,0.,0.])
            d1 = numpy.cross(direction, v)
            d1 = normaliseVector(d1)
            d2 = numpy.cross(direction, d1)
            d2 = normaliseVector(d2)
    
            E_vector = numpy.cross(self.E_vector, direction)
            E_vector = numpy.cross(E_vector, direction)
    
            ray_data = numpy.zeros((2*count)+1, dtype=ray_dtype)
            
            theta0 = self.wavelength/(numpy.pi*radius*1000.0)
            lr = numpy.array([1,-1])[:,None,None]
            eye = numpy.array([1,1])[:,None,None]
            alpha = (numpy.arange(count)*(2*numpy.pi/count))[None,:,None]
            origins = eye*radius*(d1*numpy.sin(alpha) + d2*numpy.cos(alpha))
            directions = theta0*lr*(d1*numpy.cos(alpha) - d2*numpy.sin(alpha))
            
            origins.shape = (-1,3)
            directions.shape = (-1,3)
            directions += direction
            directions = normaliseVector(directions)
            origins = origins - (self.working_distance*directions)
                
            ray_data['origin'][1:] = origins
            ray_data['origin'] += origin + (self.working_distance*direction)
            ray_data['direction'][1:] = directions
            ray_data['direction'][0] = direction
            ray_data['wavelength_idx'] = 0
            ray_data['E_vector'] = [normaliseVector(E_vector)]
            ray_data['E1_amp'] = self.E1_amp
            ray_data['E2_amp'] = self.E2_amp
            ray_data['refractive_index'] = 1.0+0.0j
            ray_data['normal'] = [[0,1,0]]
            rays = RayCollection.from_array(ray_data)
            rays.wavelengths = numpy.asarray(self.wavelength_list)
        except:
            import traceback
            traceback.print_exc()
            return RayCollection(5)
        return rays
    

class ConfocalRaySource(SingleRaySource):
    abstract = False
    focus = TupleVector
    direction = UnitTupleVector
    number = Range(1,50,20, editor_traits={'mode':'spinner'})
    theta = Range(0.0,90.0,value=30)
    working_dist = Float(100.0,editor=NumEditor)
    rings = Range(0,50,3, editor_traits={'mode':'spinner'})
    reverse = 1    #-1 is converging, +1 is diverging
    principle_axes = Property(Tuple(Array,Array), depends_on="direction")
    
    #view_ray_ids = numpy.arange(20)
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="focus, direction, number, rings, theta, working_dist, max_ray_len")
    
    geom_grp = VGroup(Group(Item('focus', show_label=False,resizable=True), 
                            show_border=True,
                            label="Focus position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item('number'),
                       Item('rings'),
                       Item('theta', label="angle"),
                       Item('working_dist', label="working dist"),
                       label="Geometry")
    
    @cached_property
    def _get_principle_axes(self):
        direction = self.direction
        max_axis = numpy.abs(direction).argmax()
        if max_axis==0:
            v = numpy.array([0.,1.,0.])
        else:
            v = numpy.array([1.,0.,0.])
        d1 = numpy.cross(direction, v)
        d1 = normaliseVector(d1)
        d2 = numpy.cross(direction, d1)
        d2 = normaliseVector(d2)
        return (d1, d2)
    
    @on_trait_change("focus, direction, number, theta, working_dist, max_ray_len, rings")
    def on_update(self):
        self.data_source.modified()
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        rev = self.reverse #to determine converging or diverging  
        focus = numpy.array(self.focus)
        direction = numpy.array(self.direction)
        working_dist = self.working_dist
        origin = focus - rev*direction*working_dist
        count = self.number
        rings = self.rings
        theta = self.theta
        radius = numpy.tan(numpy.pi * theta/180) * working_dist
        
        d1, d2 = self.principle_axes
        ray_data = numpy.zeros((rings*count)+1, dtype=ray_dtype)
        
        radii = ((numpy.arange(rings)+1)*(radius/rings))[:,None,None]
        angles = (numpy.arange(count)*(2*numpy.pi/count))[None,:,None]
        offsets = radii*(d1*numpy.sin(angles) + d2*numpy.cos(angles)) 
        offsets.shape = (-1,3)

        ray_data['origin'][1:] = offsets
        ray_data['origin'] += origin
        ray_data['direction'][0] = normaliseVector(rev*direction)
        ray_data['direction'][1:] = normaliseVector((rev*direction*working_dist) - offsets)
        ray_data['length'] = self.max_ray_len
        ray_data['E_vector'] = self.E_vector
        ray_data['E1_amp'] = self.E1_amp
        ray_data['E2_amp'] = self.E2_amp
        ray_data['wavelength_idx'] = 0
        ray_data['refractive_index'] = 1.0
        ray_data['normal'] = [0,1,0]
        rays = RayCollection.from_array(ray_data)
        return rays
    
    @cached_property
    def _get_InputDetailRays(self):
        focus = numpy.array(self.focus)
        direction = numpy.array(self.direction)
        working_dist = self.working_dist
        origin = focus - direction*working_dist
        count = self.number
        theta = self.theta
        radius = numpy.tan(numpy.pi * theta/180)
        
        d1, d2 = self.principle_axes
        
        det_res = self.detail_resolution
        slen = det_res * 1j
        newaxis = numpy.newaxis
        
        d1o, d2o = numpy.ogrid[-radius:radius:slen,-radius:radius:slen]
        
        index = numpy.arange(det_res*det_res).reshape(det_res,det_res)
        
        cells = numpy.dstack([index[:-1,:-1],
                                    index[1:,:-1],
                                    index[1:,1:],
                                    index[:-1,1:]]).reshape(-1,4)
        
        offsets = (d1o[:,:,newaxis]*d1[newaxis,newaxis,:] \
                    + d2o[:,:,newaxis]*d2[newaxis,newaxis,:]).reshape(-1,3)
                    
        in_rad = (offsets**2).sum(axis=-1) < radius**2
        
        #cull points outside the radius
        offsets = offsets[in_rad,:]
        
        #now cull cells with points outside the radius
        map = numpy.ones(in_rad.shape, numpy.int) * -1
        map[in_rad] = numpy.arange(in_rad.sum())
        
        mask = in_rad[cells].all(axis=-1)
        cells = map[cells[mask,:]]

        origins = origin[newaxis,:] + offsets*working_dist
        directions = normaliseVector(direction[newaxis,:] - offsets)
        
        rays = RayCollection(origin=origins, direction=directions,
                             max_length=self.max_ray_len)
        rays.set_polarisation(1, 0, 0)
        size = origins.shape[0],1
        rays.E1_amp = numpy.ones(size, dtype=numpy.complex128)
        rays.E2_amp = numpy.zeros(size, dtype=numpy.complex128)
        rays.refractive_index = numpy.ones(size, dtype=numpy.complex128)
        rays.cells = cells
        rays.offset_length = numpy.sqrt(((origins - focus)**2).sum(axis=-1)).reshape(-1,1)
        #print cells.max(), rays.number, "check"
        return rays


class AdHocSource(BaseRaySource):
    '''create a source by specifying the input rays yourself''' 
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, number, rings, radius, max_ray_len")

    def __init__(self,*args,**kwargs):
        self.rays = kwargs.pop('rays')
        super(AdHocSource,self).__init__(*args,**kwargs)

    @cached_property
    def _get_input_rays(self):
        #i actually don't understand traits well at all.  this whole class is a hack.
        return self.rays
    
    
class RayFieldSource(SingleRaySource):    
    show_mesh = Bool(True)
    mesh_source = Instance(tvtk.ProgrammableSource, transient=True)
    mesh_mapper = Instance(tvtk.PolyDataMapper, (), transient=True)
    mesh_actor = Instance(tvtk.Actor, transient=True)
    
    def _get_actors(self):
        actors = [self.ray_actor, self.start_actor, self.normals_actor, self.mesh_actor]
        return actors
        
    def _mesh_source_default(self):
        source = tvtk.ProgrammableSource()
        def execute():
            if not self.show_mesh:
                return
            output = source.poly_data_output
            rays = self.input_rays
            points = rays.origin
            neighbours = rays.neighbours
            fz = frozenset
            tris = set()
            for i in range(points.shape[0]):
                nb = neighbours[i]
                if any(a<=0 for a in nb):
                    continue
                nb1, nb2, nb3, nb4, nb5, nb6 = nb
                these = ((i, nb1, nb2),
                             (i, nb2, nb3),
                             (i, nb3, nb4),
                             (i, nb4, nb5),
                             (i, nb5, nb6),
                             (i, nb6, nb1))
                for tri in these:
                    tris.add( fz(tri) )
            output.points = points
            output.polys = [list(a) for a in tris]
            
            E1 = rays.E1_amp
            E2 = rays.E2_amp
            I = E1.real**2 + E1.imag**2 + E2.real**2 + E2.imag**2
            output.point_data.scalars = I
            self.mesh_mapper.scalar_range = (I.min(), I.max())

        source.set_execute_method(execute)
        return source
    
    def _mesh_actor_default(self):
        source = self.mesh_source
        map = self.mesh_mapper
        map.input_connection = source.output_port
        act = tvtk.Actor(mapper=map)
        #act.property.color = (0.0,1.0,0.0)
        act.property.representation = "surface"
        return act
    
    @on_trait_change("origin, direction")
    def _on_update_mesh(self):
        self.mesh_source.modified()
    

class HexagonalRayFieldSource(RayFieldSource):
    abstract = False
    radius = Float(10.,editor=NumEditor)
    resolution = Float(10.0, editor=NumEditor)
    gauss_width = Float(2.0, editor=NumEditor)
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, resolution, wavelength, gauss_width, radius, max_ray_len, E_vector")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item('radius'),
                       Item('resolution'),
                       Item('gauss_width'),
                       label="Geometry")

    @on_trait_change("direction, spacing, radius, max_ray_len")
    def on_update(self):
        self.data_source.modified()
        self.mesh_source.modified()
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        origin = numpy.array(self.origin)
        direction = numpy.array(self.direction)
        radius = self.radius
        spacing = radius/self.resolution
        max_axis = numpy.abs(direction).argmax()
        if max_axis==0:
            v = numpy.array([0.,1.,0.])
        else:
            v = numpy.array([1.,0.,0.])
        d1 = numpy.cross(direction, v)
        d1 = normaliseVector(d1)
        d2 = numpy.cross(direction, d1)
        d2 = normaliseVector(d2)

        E_vector = numpy.cross(self.E_vector, direction)
        E_vector = numpy.cross(E_vector, direction)
        
        cosV = numpy.cos(numpy.pi*30/180.)
        sinV = numpy.sin(numpy.pi*30/180.)
        d2 = cosV*d2 + sinV*d1
        
        nsteps = int(radius/(spacing*numpy.cos(numpy.pi*30/180.)))
        i = numpy.arange(-nsteps-1, nsteps+2)
        j = numpy.arange(-nsteps-1, nsteps+2)
        
        vi, vj = numpy.meshgrid(i,j)
        label = numpy.arange(vi.size).reshape(*vi.shape) #give each point a unique index
        neighbours = numpy.full((label.shape[0], label.shape[1], 6), -1, 'i')
        neighbours[1:,:,3] = label[:-1,:]
        neighbours[:-1,:,0] = label[1:,:]
        neighbours[:,1:,4] = label[:,:-1]
        neighbours[:,:-1,1] = label[:,1:]
        neighbours[1:,:-1,2] = label[:-1,1:]
        neighbours[:-1,1:,5] = label[1:,:-1]
        
        neighbours.shape = -1,6
        
        vi.shape = -1
        vj.shape = -1
        
        ri = ((vi + sinV*vj)**2 + (cosV*vj**2))
        select = ri < (radius/spacing)**2
        
        xi = vi[select]
        yj = vj[select]
        
        backmap = numpy.full(vi.shape[0]+1, -1, 'i')
        backmap[:-1][select] = numpy.arange(len(xi))
        selnb = backmap[neighbours]
        neighbours = selnb[select,:]
        
        offsets = xi[:,None]*d1 + yj[:,None]*d2
        offsets *= spacing
        
        gauss = numpy.exp(-(ri[select]*(spacing**2)/((self.gauss_width)**2))) 
        
        ray_data = numpy.zeros(offsets.shape[0], dtype=ray_dtype)
            
        ray_data['origin'] = offsets
        ray_data['origin'] += origin
        ray_data['direction'] = direction
        ray_data['wavelength_idx'] = 0
        ray_data['E_vector'] = [normaliseVector(E_vector)]
        ray_data['E1_amp'] = self.E1_amp * gauss
        ray_data['E2_amp'] = self.E2_amp * gauss
        ray_data['refractive_index'] = 1.0+0.0j
        ray_data['normal'] = [[0,1,0]]
        rays = RayCollection.from_array(ray_data)
        rays.neighbours = neighbours
        return rays
    
    
class ConfocalRayFieldSource(HexagonalRayFieldSource):
    abstract = False
    angle = Float(10.0)
    
    ### Number of angle steps along the radius
    resolution = Range(0.0, None, 10.0)
    
    working_dist = Float(100.0)
    
    numerical_aperture = Float(0.12)
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, angle, resolution, "
                                    "working_dist, max_ray_len, E_vector, wavelength, "
                                    "E1_amp, E2_amp, numerical_aperture")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item('wavelength', editor=NumEditor),
                       Item('numerical_aperture', editor=NumEditor),
                       Item('angle', editor=NumEditor),
                       Item('resolution', editor=NumEditor),
                       Item('working_dist', editor=NumEditor),
                       label="Geometry")

    @on_trait_change("angle, resolution, numerical_aperture, working_dist, max_ray_len")
    def on_update(self):
        self.data_source.modified()
        self.mesh_source.modified()
    
    @cached_property
    def _get_input_rays(self):
        try:
            origin = numpy.array(self.origin)
            direction = numpy.array(self.direction)
            
            radius = numpy.tan(self.angle*numpy.pi/180)
            spacing = radius /( self.resolution + 0.5 )
            
            max_axis = numpy.abs(direction).argmax()
            if max_axis==0:
                v = numpy.array([0.,1.,0.])
            else:
                v = numpy.array([1.,0.,0.])
            d1 = numpy.cross(direction, v)
            d1 = normaliseVector(d1)
            d2 = numpy.cross(direction, d1)
            d2 = normaliseVector(d2)
            
            cosV = numpy.cos(numpy.pi*30/180.)
            sinV = numpy.sin(numpy.pi*30/180.)
            d2 = cosV*d2 + sinV*d1
            
            nsteps = int(radius/(spacing*numpy.cos(numpy.pi*30/180.)))
            i = numpy.arange(-nsteps-1, nsteps+2)
            j = numpy.arange(-nsteps-1, nsteps+2)
            
            vi, vj = numpy.meshgrid(i,j)
            label = numpy.arange(vi.size).reshape(*vi.shape) #give each point a unique index
            neighbours = numpy.full((label.shape[0], label.shape[1], 6), -1, 'i')
            neighbours[1:,:,3] = label[:-1,:]
            neighbours[:-1,:,0] = label[1:,:]
            neighbours[:,1:,4] = label[:,:-1]
            neighbours[:,:-1,1] = label[:,1:]
            neighbours[1:,:-1,2] = label[:-1,1:]
            neighbours[:-1,1:,5] = label[1:,:-1]
            
            neighbours.shape = -1,6
            
            vi.shape = -1
            vj.shape = -1
            
            ri = ((vi + sinV*vj)**2 + (cosV*vj**2))
            select = ri < (radius/spacing)**2
            
            xi = vi[select]
            yj = vj[select]
            
            backmap = numpy.full(vi.shape[0]+1, -1, 'i')
            backmap[:-1][select] = numpy.arange(len(xi))
            selnb = backmap[neighbours]
            neighbours = selnb[select,:]
            
            offsets = xi[:,None]*d1 + yj[:,None]*d2
            offsets *= spacing
            
            directions = direction[None,:] - offsets
            directions = normaliseVector(directions)
            
            E_vector = numpy.cross(self.E_vector, directions)
            E_vector = numpy.cross(E_vector, directions)
            E_vector = normaliseVector(E_vector)
            
            r_na = numpy.tan(numpy.arcsin(self.numerical_aperture))
            gauss = numpy.exp(-(ri[select]*(spacing**2))/(r_na**2))*spacing
            
            ray_data = numpy.zeros(offsets.shape[0], dtype=ray_dtype)
                
            wl = self.wavelength
            path_length = numpy.sqrt(((offsets + direction[None,:])**2).sum(axis=-1))*self.working_dist
            phase = -2*numpy.pi*((1000*path_length/wl))#%1.0)
            ray_data['origin'] = offsets*self.working_dist
            ray_data['origin'] += origin
            ray_data['direction'] = directions
            ray_data['wavelength_idx'] = 0
            ray_data['E_vector'] = E_vector
            ray_data['E1_amp'] = self.E1_amp * gauss
            ray_data['E2_amp'] = self.E2_amp * gauss
            ray_data['refractive_index'] = 1.0+0.0j
            ray_data['normal'] = [[0,1,0]]
            ray_data['phase'] = phase
            rays = RayCollection.from_array(ray_data)
            rays.neighbours = neighbours
        except:
            traceback.print_exc()
            return RayCollection(0)
        print("Made input rays.")
        return rays
    
    
class SingleGaussletSource(SingleRaySource):
    beam_waist = Float(100.0) #in microns
    working_distance = Float(0.0) #the distance from the beam waist to the source location
    
    input_rays = Property(Instance(RayCollection), 
                         depends_on="origin, direction, beam_waist, wavelength, "
                         "max_ray_len, E_vector, working_distance")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item("wavelength"),
                       Item('beam_waist', tooltip="1/e^2 beam intensity radius, in microns"),
                       Item('working_distance', tooltip="The distance from the source to the beam waist"),
                       label="Geometry")
    
    
    @on_trait_change("direction, beam_waist, max_ray_len, working_distance")
    def on_update(self):
        self.data_source.modified()
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        try:
            origin = numpy.array(self.origin)
            direction = numpy.array(self.direction)
            radius = self.beam_waist/2000.0 #convert to mm
            max_axis = numpy.abs(direction).argmax()
            if max_axis==0:
                v = numpy.array([0.,1.,0.])
            else:
                v = numpy.array([1.,0.,0.])
            d1 = numpy.cross(direction, v)
            d1 = normaliseVector(d1)
            d2 = numpy.cross(direction, d1)
            d2 = normaliseVector(d2)
    
            E_vector = numpy.cross(self.E_vector, direction)
            E_vector = numpy.cross(E_vector, direction)
    
            count = 3
            ray_data = numpy.zeros((2*count)+1, dtype=ray_dtype)
            
            theta0 = self.wavelength/(numpy.pi*radius*1000.0)
            lr = numpy.array([1,-1])[:,None,None]
            eye = numpy.array([1,1])[:,None,None]
            alpha = (numpy.arange(count)*(2*numpy.pi/count))[None,:,None]
            origins = eye*radius*(d1*numpy.sin(alpha) + d2*numpy.cos(alpha))
            directions = theta0*lr*(d1*numpy.cos(alpha) - d2*numpy.sin(alpha))
            
            origins.shape = (-1,3)
            directions.shape = (-1,3)
            directions += direction
            directions = normaliseVector(directions)
            origins = origins - (self.working_distance*directions)
                
            ray_data['origin'] = origin
            ray_data['origin'][1:] += origins + (self.working_distance*direction)
            ray_data['direction'][1:] = directions
            ray_data['direction'][0] = direction
            ray_data['wavelength_idx'] = 0
            ray_data['E_vector'] = [normaliseVector(E_vector)]
            ray_data['E1_amp'] = self.E1_amp
            ray_data['E2_amp'] = self.E2_amp
            ray_data['refractive_index'] = 1.0+0.0j
            ray_data['normal'] = [[0,1,0]]
            ray_data['ray_type_id'] = GAUSSLET_
            ray_data['ray_type_id'][1:] |= PARABASAL_ 
            rays = RayCollection.from_array(ray_data)
            
            neighbours = numpy.full((7, 6), -1, 'i')
            neighbours[0,:] = [1,2,3,4,5,6]
            rays.neighbours = neighbours
        except:
            import traceback
            traceback.print_exc()
            return RayCollection(5)
        return rays
    
    



from raypier.tracer import RayTraceModel
from raypier.paraxial import ParaxialLens
from raypier.sources import ConfocalRaySource

lens = ParaxialLens(diameter = 50.0,
                    focal_length = 50.0,
                    centre = (0,0,0),
                    direction = (1,0,0))

lens2 = ParaxialLens(diameter = 50.0,
                    focal_length = 50.0,
                    centre = (50.0,0,0),
                    direction = (1,0,0))

source = ConfocalRaySource(focus = (-50.0,0,0),
                           direction = (1,0,0),
                           working_dist= 10.0,
                           theta=15.0)

model = RayTraceModel(optics=[lens,lens2],
                      sources=[source])

model.configure_traits()



'''
Support for achromat lenses 

Created on 23 Nov 2018

:author: bryan
'''
from traits.api import Float, Instance, on_trait_change
from traitsui.api import View, Item, VGroup

from raypier.bases import NumEditor, Traceable
from raypier.lenses import BaseLens
from raypier.dispersion import BaseDispersionCurve, NamedDispersionCurve, \
        NondispersiveCurve
from raypier.core.cmaterials import CoatedDispersiveMaterial
from raypier.core.cfaces import SphericalFace
from raypier.core.ctracer import FaceList
from raypier.vtk_algorithms import EmptyGridSource

from tvtk.api import tvtk
from numpy import sqrt


class Singlet(BaseLens):
    """
    A dispersive singlet lens with spherical surfaces. Not really achromatic at all.
    """
    abstract = False
    name = "sperical singlet"
    
    #: Lens diameter in mm
    diameter = Float(25.0)
    
    #: Center thickness in mm
    CT = Float(6.0, desc="first centre thickness")
    
    #: Radius of curvature for the first surface, in mm
    curvature1 = Float(43.96, desc="radius of curvature for first outer face")
    
    #: Radius of curvature for the second surface, in mm
    curvature2 = Float(-42.90, desc="radius of curvature for the inner face")
    
    #: Dispersion curve for the lens bulk :type raypier.core.cmaterials.BaseDispersionCurve:
    dispersion = Instance(BaseDispersionCurve, desc="Dispersion curve for the glass")
    
    #: Dispersion curve for the lens coating :type raypier.core.cmaterials.BaseDispersionCurve:
    dispersion_coating = Instance(BaseDispersionCurve, desc="Dispersion curve for the coating")
    
    #: Coating thickness, in microns
    coating_thickness = Float(0.25, desc="Thickness of the AR coating, in microns")
    
    _material1 = Instance(CoatedDispersiveMaterial, ())
    _material2 = Instance(CoatedDispersiveMaterial, ())
    
    vtk_grid = Instance(EmptyGridSource, ())
    vtk_cylinder = Instance(tvtk.Cylinder, ())
    vtk_sphere1 = Instance(tvtk.Sphere, ())
    vtk_sphere2 = Instance(tvtk.Sphere, ())
    
    clip2 = Instance(tvtk.ClipDataSet, ())
    clip3 = Instance(tvtk.ClipDataSet, ())
    
    traits_view = View(VGroup(
                       Traceable.uigroup,  
                       Item('CT', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       Item('curvature1', editor=NumEditor),
                       Item('curvature2', editor=NumEditor),
                       )
                    )
    
    @on_trait_change("dispersion, dispersion_coating, coating_thickness")
    def on_materials_changed(self):
        self.apply_materials()
        self.update = True
    
    def apply_materials(self):
        air = NondispersiveCurve(1.0)
        self._material1.dispersion_inside = self.dispersion
        self._material1.dispersion_outside = air
        self._material1.dispersion_coating = self.dispersion_coating
        self._material1.coating_thickness = self.coating_thickness
        
        self._material2.dispersion_inside = air
        self._material2.dispersion_outside = self.dispersion
        self._material2.dispersion_coating = self.dispersion_coating
        self._material2.coating_thickness = self.coating_thickness
        
    @on_trait_change("diameter, CT, curvature1, curvature2")
    def on_geometry_changed(self):
        face1, face2 = self.faces.faces
        face1.diameter = self.diameter
        face1.z_height = self.CT/2
        face1.curvature = self.curvature1
        
        face2.diameter = self.diameter
        face2.z_height = -self.CT/2
        face2.curvature = self.curvature2
        
        self.update = True
    
    def _faces_default(self):
        self.apply_materials()
        fl = FaceList(owner=self)
        fl.faces = [ 
                  SphericalFace(owner=self, 
                                diameter=self.diameter,
                                material=self._material1,
                                z_height=self.CT/2, 
                                curvature=self.curvature1),
                  SphericalFace(owner=self, 
                                diameter=self.diameter,
                                material=self._material2,
                                z_height=-self.CT/2, 
                                curvature=self.curvature2),
                ]
        return fl
    
    @on_trait_change("CT, diameter, curvature1, curvature2")
    def config_pipeline(self):
        ct = self.CT
        rad = self.diameter/2
        curve1 = self.curvature1
        curve2 = self.curvature2
        
        size = 41
        spacing = 2*rad / (size-1)
        if curve1 >= 0.0:
            extra1=0.0
        else:
            extra1 = -curve1 - sqrt(curve1**2 - rad**2)
            
        if curve2 <= 0.0:
            extra2=0.0
        else:
            extra2 = curve2 - sqrt(curve2**2 - rad**2)
            
        lsize1 = int((ct/2+extra1)/spacing) + 2
        lsize2 = int((ct/2+extra2)/spacing) + 2
        
        grid = self.vtk_grid
        grid.dimensions = (size,size,lsize1+lsize2)
        grid.origin = (-rad, -rad, -lsize2*spacing)
        grid.spacing = (spacing, spacing, spacing)
                      
        cyl = self.vtk_cylinder
        cyl.center = (0,0,0)
        cyl.radius = rad
        
        s1 = self.vtk_sphere1
        s1.center = (0,0,ct/2 - curve1)
        s1.radius = abs(curve1)
        
        s2 = self.vtk_sphere2
        s2.center = (0,0,-ct/2-curve2)
        s2.radius = abs(curve2)
        
        self.clip2.inside_out = bool(curve1 >= 0.0)
        self.clip3.inside_out = bool(curve2 <= 0.0)
        
        self.vtk_grid.modified()
        self.update=True
        
                                         
    def _pipeline_default(self):
        grid = self.vtk_grid
        #grid.set_execute_method(self.create_grid)
        grid.modified()
        
        trans = tvtk.Transform()
        trans.rotate_x(90.)
        cyl = self.vtk_cylinder
        cyl.transform = trans
        
        clip1 = tvtk.ClipVolume(input_connection=grid.output_port,
                                 clip_function=self.vtk_cylinder,
                                 inside_out=1)
        
        self.clip2.set(input_connection = clip1.output_port,
                      clip_function=self.vtk_sphere1,
                      inside_out=1)
        
        self.clip3.set(input_connection = self.clip2.output_port,
                       clip_function=self.vtk_sphere2,
                       inside_out=1)
        
        topoly = tvtk.GeometryFilter(input_connection=self.clip3.output_port)
        
        norms = tvtk.PolyDataNormals(input_connection=topoly.output_port)
        
        transF = tvtk.TransformFilter(input_connection=norms.output_port, 
                                      transform=self.transform)
        
        self.on_geometry_changed()
        self.config_pipeline()
        grid.modified()
        return transF
    

class Doublet(BaseLens):
    abstract = False
    name = "Achromatic Doublet"
    
    CT1 = Float(6.0, desc="first centre thickness")
    CT2 = Float(4.0, desc="second centre thickness")
    diameter = Float(25.0)
    curvature1 = Float(43.96, desc="radius of curvature for first outer face")
    curvature2 = Float(-42.90, desc="radius of curvature for the inner face")
    curvature3 = Float(-392.21, desc="radius of curvature for the second outer face")
    dispersion1 = Instance(BaseDispersionCurve, desc="Dispersion curve for first material")
    dispersion2 = Instance(BaseDispersionCurve, desc="Dispersion curve for second material")
    dispersion_coating = Instance(BaseDispersionCurve, desc="Dispersion curve for the coating")
    coating_thickness = Float(0.25, desc="Thickness of the AR coating, in microns")
    
    _material1 = Instance(CoatedDispersiveMaterial, ())
    _material2 = Instance(CoatedDispersiveMaterial, ())
    _material3 = Instance(CoatedDispersiveMaterial, ())
    
    vtk_grid = Instance(EmptyGridSource, ())
    vtk_cylinder = Instance(tvtk.Cylinder, ())
    vtk_sphere1 = Instance(tvtk.Sphere, ())
    vtk_sphere2 = Instance(tvtk.Sphere, ())
    vtk_sphere3 = Instance(tvtk.Sphere, ())
    
    clip2 = Instance(tvtk.ClipDataSet, ())
    clip3 = Instance(tvtk.ClipDataSet, ())
    
    traits_view = View(VGroup(
                       Traceable.uigroup,  
                       Item('CT1', editor=NumEditor),
                       Item('CT2', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       Item('curvature1', editor=NumEditor),
                       Item('curvature2', editor=NumEditor),
                       Item('curvature3', editor=NumEditor),
                       )
                    )
    
    @on_trait_change("dispersion1, dispersion2, dispersion_coating, coating_thickness")
    def on_materials_changed(self):
        self.apply_materials()
        self.update = True
    
    def apply_materials(self):
        air = NondispersiveCurve(1.0)
        self._material1.dispersion_inside = self.dispersion1
        self._material1.dispersion_outside = air
        self._material1.dispersion_coating = self.dispersion_coating
        self._material1.coating_thickness = self.coating_thickness
        
        self._material2.dispersion_inside = self.dispersion2
        self._material2.dispersion_outside = self.dispersion1
        self._material2.dispersion_coating = self.dispersion_coating
        self._material2.coating_thickness = 0.0
        
        ### We reverse inside and outside because really, outside means
        ### "the direction the surface normals point".
        self._material3.dispersion_inside = air
        self._material3.dispersion_outside = self.dispersion2
        self._material3.dispersion_coating = self.dispersion_coating
        self._material3.coating_thickness = self.coating_thickness
        
    @on_trait_change("diameter, CT1, CT2, curvature1, curvature2, curvature3")
    def on_geometry_changed(self):
        face1, face2, face3 = self.faces.faces
        face1.diameter = self.diameter
        face1.z_height = self.CT1
        face1.curvature = self.curvature1
        
        face2.diameter = self.diameter
        face2.curvature = self.curvature2
        
        face3.diameter = self.diameter
        face3.z_height = -self.CT2
        face3.curvature = self.curvature3
        self.update = True
    
    def _faces_default(self):
        self.apply_materials()
        fl = FaceList(owner=self)
        fl.faces = [ 
                  SphericalFace(owner=self, 
                                diameter=self.diameter,
                                material=self._material1,
                                z_height=self.CT1, 
                                curvature=self.curvature1),
                  SphericalFace(owner=self, 
                                diameter=self.diameter,
                                material=self._material2,
                                z_height=0.0, 
                                curvature=self.curvature2),
                  SphericalFace(owner=self, 
                                diameter=self.diameter,
                                material=self._material3,
                                z_height=-self.CT2, 
                                curvature=self.curvature3),
                ]
        return fl
    
    @on_trait_change("CT1, CT2, diameter, curvature1, curvature3, curvature2")
    def config_pipeline(self):
        ct1 = self.CT1
        ct2 = self.CT2
        rad = self.diameter/2
        curve1 = self.curvature1
        curve2 = self.curvature2
        curve3 = self.curvature3
        
        size = 41
        spacing = 2*rad / (size-1)
        if curve1 >= 0.0:
            extra1=0.0
        else:
            extra1 = -curve1 - sqrt(curve1**2 - rad**2)
            
        if curve3 <= 0.0:
            extra3=0.0
        else:
            extra3 = curve3 - sqrt(curve3**2 - rad**2)
            
        lsize1 = int((ct1+extra1)/spacing) + 2
        lsize2 = int((ct2+extra3)/spacing) + 2
        
        grid = self.vtk_grid
        grid.dimensions = (size,size,lsize1+lsize2)
        grid.origin = (-rad, -rad, -lsize2*spacing)
        grid.spacing = (spacing, spacing, spacing)
                      
        cyl = self.vtk_cylinder
        cyl.center = (0,0,0)
        cyl.radius = rad
        
        s1 = self.vtk_sphere1
        s1.center = (0,0,ct1 - curve1)
        s1.radius = abs(curve1)
        
        s2 = self.vtk_sphere2
        s2.center = (0,0,-curve2)
        s2.radius = abs(curve2)
        
        s3 = self.vtk_sphere3
        s3.center = (0,0,-ct2 - curve3)
        s3.radius = abs(curve3)
        
        self.clip2.inside_out = bool(curve1 >= 0.0)
        self.clip3.inside_out = bool(curve3 <= 0.0)
        
        self.vtk_grid.modified()
        self.update=True
        
                                         
    def _pipeline_default(self):
        grid = self.vtk_grid
        #grid.set_execute_method(self.create_grid)
        grid.modified()
        
        trans = tvtk.Transform()
        trans.rotate_x(90.)
        cyl = self.vtk_cylinder
        cyl.transform = trans
        
        clip1 = tvtk.ClipVolume(input_connection=grid.output_port,
                                 clip_function=self.vtk_cylinder,
                                 inside_out=1)
        
        self.clip2.set(input_connection = clip1.output_port,
                      clip_function=self.vtk_sphere1,
                      inside_out=1)
        
        self.clip3.set(input_connection = self.clip2.output_port,
                       clip_function=self.vtk_sphere3,
                       inside_out=1)
        
        clip4 = tvtk.ClipDataSet(input_connection = self.clip3.output_port,
                       clip_function=self.vtk_sphere2,
                       inside_out=1)
        
        clip5 = tvtk.ClipDataSet(input_connection = self.clip3.output_port,
                       clip_function=self.vtk_sphere2,
                       inside_out=0)
        
        topoly = tvtk.GeometryFilter(input_connection=clip4.output_port)
        topoly2 = tvtk.GeometryFilter(input_connection=clip5.output_port)
        
        append = tvtk.AppendPolyData()
        append.add_input_connection(topoly.output_port)
        append.add_input_connection(topoly2.output_port)
        
        norms = tvtk.PolyDataNormals(input_connection=append.output_port)
        
        transF = tvtk.TransformFilter(input_connection=norms.output_port, 
                                      transform=self.transform)
        
        self.on_geometry_changed()
        self.config_pipeline()
        grid.modified()
        return transF
    
    def make_step_shape(self):
        """Creates an OpenCascade BRep Shape
        representation of the object, which can be
        exported to STEP format"""
        from raypier.step_export import make_spherical_lens2, make_compound
        centre = self.centre
        direction = self.direction
        x_axis = self.x_axis
        diameter = self.diameter
        CT1 = self.CT1
        CT2 = 0.0
        CT3 = -self.CT2
        curvature1 = self.curvature1
        curvature2 = self.curvature2
        curvature3 = self.curvature3
        
        lens1 = make_spherical_lens2(CT1, CT2, diameter, curvature1, 
                                     curvature2, centre, direction, x_axis)
        lens2 = make_spherical_lens2(CT2, CT3, diameter, curvature2, 
                                     curvature3, centre, direction, x_axis)
        doublet = make_compound([lens1, lens2])
        return doublet, "yellow"
    
    
class EdmundOptic45805(Doublet):
    name = "Edmund Optics Achromat#45805"
    CT1 = 6.0
    CT2 = 4.0
    diameter = 25.4
    offset = 0.0
    curvature1 = 43.96
    curvature2 = -42.90
    curvature3 = -392.21
    dispersion1 = NamedDispersionCurve("N-LAK22")
    dispersion2 = NamedDispersionCurve("N-SF6")
    dispersion_coating = NondispersiveCurve(refractive_index=1.37)
    coating_thickness = 0.283 #microns
    
    
class EdmundOptic45806(Doublet):
    name = "Edmund Optics Achromat#45-806"
    CT1 = 6.0
    CT2 = 4.0
    diameter = 25.4
    offset = 0.0
    curvature1 = 57.01
    curvature2 = -56.72
    curvature3 = -656.51
    dispersion1 = NamedDispersionCurve("N-LAK22")
    dispersion2 = NamedDispersionCurve("N-SF6")
    dispersion_coating = NondispersiveCurve(refractive_index=1.37)
    coating_thickness = 0.283 #microns
    



"""
Functions relating to the evaluation of the optical E-field by
summation of General Astigmatic Gaussian Beams
"""


from .cfields import sum_gaussian_modes, evaluate_modes as evaluate_modes_c
from .utils import normaliseVector, dotprod
from .ctracer import RayCollection

import numpy
scipy = None



def find_ray_gen(probe_centre, traced_rays):
    probe = numpy.asarray(probe_centre)
    waypoints = [rays.origin.mean(axis=0) for rays in traced_rays]
    waypoints.append(traced_rays[-1].termination.mean(axis=0))
    
    dist = [ ((wpt-probe)**2).sum() for wpt in waypoints ]
    
    imin = numpy.argmin(dist)
    if imin + 1 >= len(waypoints):
        return -1
    
    closest = waypoints[imin]
    
    dir1 = closest - waypoints[imin-1]
    dir1 = normaliseVector(dir1)
    
    dir2 = waypoints[imin+1] - closest
    dir2 = normaliseVector(dir2)
    
    normal = normaliseVector(dir1 + dir2)
    if dotprod(probe-closest, normal) > 0:
        idx = imin
    else:
        idx = imin-1
    if idx < 0:
        idx = 0
    if idx > len(waypoints)-2:
        idx = -1
    return idx
    
    
    ### unfinished ###


def project_to_sphere(rays, centre=(0,0,0), radius=10.0):
    """project the given set of rays back to their intercept 
    with a sphere at the given centre and radius.
    
    rays - an array of ray_t dtype
    """
    centre = numpy.asarray(centre).reshape(1,3)
    origin = rays['origin'] - centre
    direction = rays['direction'] #assume this is already normalised
    c = (origin**2).sum(axis=1) - (radius**2)
    b = 2*(direction * origin).sum(axis=1)
    a = 1 
    
    d = b**2 - 4*c
    selector = (d>=0.0)
    d = numpy.sqrt(d[selector])
    b = b[selector]
    root1 = (-b + d)/2
    root2 = (-b - d)/2
    #choose most negative path to intersection
    alpha = numpy.column_stack((root1, root2)).min(axis=1)
    
    rays = rays[selector]
    rays['origin'] += alpha[:,None]*direction[selector]
    
    rays['accumulated_path'] += alpha*rays['refractive_index'].real
    return rays 
    

def evaluate_neighbours(rays, neighbours_idx):
    """For each of a rays neighbours, we need to project the neighbour back
    onto the plane containing the main rays origin, to obtain the (x,y)
    cordinate for the neighbour ray, relative to the main ray origin
    
    rays - a ray_t array of length N
    neighbours_idx - a array on ints of shape (N,6)
    
    returns - a 5-tuple (rays, x,y,dx,dy) where x,y are N*6 arrays for the coordinate of each neighbour.
              dx and dy represent the change in direction of the neighbouring rays (i.e. 
              curvature of the wavefront). The returns rays are the subset of the input rays with
              6 neighbours (i.e. edge-rays are dropped).
              """
    
    mask = (neighbours_idx>=0).all(axis=1)
    nb = neighbours_idx[mask,:]
    n_origin = rays['origin'][nb,:]
    n_direction = rays['direction'][nb,:]
    origin = rays['origin'][mask,None,:]
    direction = rays['direction'][mask,None,:]
    E = rays['E_vector'][mask,None,:] #Assume normalised
    H = numpy.cross(E, direction) #Should also be unit length
    offsets = n_origin - origin
    alpha = -(offsets * direction).sum(axis=-1)/(n_direction * direction).sum(axis=-1)
    projected = (offsets + alpha[:,:,None]*n_direction)
    x = (projected * E).sum(axis=-1)
    y = (projected * H).sum(axis=-1)
    
    dz = (n_direction * direction).sum(axis=-1)
    dx = (n_direction * E).sum(axis=-1)/dz 
    dy = (n_direction * H).sum(axis=-1)/dz
    return (rays[mask], x,y,dx,dy)


def evaluate_neighbours_gc(gausslets):
    ga=gausslets
    origin = ga['base_ray']['origin'][:,None,:]
    direction = ga['base_ray']['direction'][:,None,:]
    n_direction = ga['para_rays']['direction']
    E = ga['base_ray']['E_vector'][:,None,:]
    H = numpy.cross(E, direction)
    
    offsets = ga['para_rays']['origin'] - origin
    #alpha = -(offsets * direction).sum(axis=-1)/(n_direction * direction).sum(axis=-1)
    ### Don't need to project onto origin normal plane as all para-rays start on this plane.
    x = (offsets * E).sum(axis=-1)
    y = (offsets * H).sum(axis=-1)
    dz = (n_direction * direction).sum(axis=-1)
    dx = (n_direction * E).sum(axis=-1)/dz 
    dy = (n_direction * H).sum(axis=-1)/dz
    
    ###This doesn't help
    #     r = x**2 + y**2
    #     r2 = (dx*x + dy*y)/r
    #     dx2 = x*r2
    #     dy2 = y*r2
    return (ga['base_ray'], x, y, dx, dy)
    
    
def evaluate_modes(neighbour_x, neighbour_y, dx, dy, blending=1.0):
    """For each ray in rays, use its nearest neighbours
    to compute the best fit for the Astigmatic Gaussian Beam
    parameters.
    
    rays - an array of ray_t with shape (N,)
    x,y,dx,dy - array of shape (N,6)
    
    For N rays, return a Nx3 complex array of coeffs"""
    global scipy
    if scipy is None:
        global lsqr
        global block_diag
        import scipy
        from scipy.sparse.linalg import lsqr
        from scipy.sparse import block_diag
    
    ### Do linear least squares on each ray and neighbours
    
    x = neighbour_x
    y = neighbour_y
    
    M = block_diag(numpy.dstack((x**2, 2*x*y, y**2)))
    b = numpy.ones(M.shape[0])
    fit = lsqr(M,b)
    im_coefs = fit[0].reshape(x.shape[0],3)

    x = neighbour_x
    y = neighbour_y
    O = numpy.zeros_like(x)
    M_ = numpy.dstack((x,y,O, O,x,y)).reshape(-1,12,3)
    b = numpy.dstack( (dx,dy) ).reshape(-1)
    M = block_diag(M_)
    fit = lsqr(M,b)
    re_coefs = fit[0].reshape(x.shape[0],3)
    
    return ((blending*1j)*im_coefs) + re_coefs


def Gamma(z, A, B, C):
    """Used in testing"""
    detG0 = (A*C) - (B*B)
    denom = (1 + (z*(A+C)) + (z*z)*detG0) #*2
    AA = (A + z*detG0)/denom
    BB = B/denom
    CC = (C + z*detG0)/denom
    return AA, BB, CC


def ExtractGamma(gausslet_collection, blending=1.0):
    """Used in Testing"""
    gc = gausslet_collection.copy_as_array() 
    rays, x, y, dx, dy = evaluate_neighbours_gc(gc)
    modes = evaluate_modes_c(x, y, dx, dy, blending=blending)
    return modes

    
def eval_Efield_from_rays(ray_collection, points, wavelengths, 
                          blending=1.0,
                          exit_pupil_offset=0.0, 
                          exit_pupil_centre=(0.0,0.0,0.0)):
    rays = ray_collection.copy_as_array() 
    radius = exit_pupil_offset
        
    if radius:
        projected = project_to_sphere(rays, exit_pupil_centre, radius)
    else:
        projected = rays
    
    neighbours_idx = ray_collection.neighbours
    rays, x, y, dx, dy = evaluate_neighbours(projected, neighbours_idx)

    modes = evaluate_modes_c(x, y, dx, dy, blending=blending)
    
    _rays = RayCollection.from_array(rays)
    
    E = sum_gaussian_modes(_rays, modes, wavelengths, points)
    
    return E


class EFieldSummation(object):
    """
    For situations where you wish to evaluate the E-field from a set of Gausslets with different sets of evaluation points,
    this class provides a small optimisation by performing the maths to convert ray-intercepts to Gaussian mode parameters
    up front.
    """
    def __init__(self, gausslet_collection, wavelengths=None, blending=1.0 ):
        if wavelengths is None:
            wavelengths = numpy.asarray(gausslet_collection.wavelengths)
        if wavelengths is None:
            raise ValueError("No wavelengths supplied")
        self.wavelengths = wavelengths
        self.gc = gc = gausslet_collection.copy_as_array() 
        rays, x, y, dx, dy = evaluate_neighbours_gc(gc)
        self.modes = evaluate_modes_c(x, y, dx, dy, blending=blending)
        self.base_rays = RayCollection.from_array(rays)
        
    def evaluate(self, points):
        """
        Called to calculate the E-field for the given points.
        """
        points = numpy.ascontiguousarray(points)
        shape = points.shape
        points.shape=(-1,3)
        E = sum_gaussian_modes(self.base_rays, 
                              self.modes, 
                              self.wavelengths, points)
        E.shape = shape
        return E


def eval_Efield_from_gausslets(gausslet_collection, points, 
                               wavelengths = None,
                               blending=1.0, **kwds):
    """
    Calculates the vector E-field is each of the points given. The returned 
    array of field-vectors will have the same length as `points` and 
    has `numpy.complex128` dtype.
    
    :param GaussletCollection gc: The set of Gausslets for which the field should be calculated
    :param ndarray[N,3] points: An array of shape (N,3) giving the points at which the field will be evaluated.
    :param ndarray[] wavelengths: A 1d array containing the wavelengths to be used for the field calculation, 
                                    overriding the wavelengths data contained by the GaussletCollection object.
    :param float blending: The 1/width of each Gaussian mode at the evaluation points. A value of unity (the default),
                            means the parabasal rays are determined to be the 1/e point in the field amplitude.
    """
    gc = gausslet_collection.copy_as_array() 
    if wavelengths is None:
        wavelengths = numpy.asarray(gausslet_collection.wavelengths)
    rays, x, y, dx, dy = evaluate_neighbours_gc(gc)
    modes = evaluate_modes_c(x, y, dx, dy, blending=blending)
    _rays = RayCollection.from_array(rays)
    E = sum_gaussian_modes(_rays, modes, wavelengths, points)
    return E




#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from traits.api import HasTraits, Array, Float, Complex,\
            Property, List, Instance, on_trait_change, Range, Any,\
            Tuple, Event, cached_property, Set, Int, Trait, Bool
from traitsui.api import View, Item, ListEditor, VSplit,\
            RangeEditor, ScrubberEditor, HSplit, VGroup, Heading
from tvtk.api import tvtk
from tvtk.pyface.scene_model import SceneModel
from tvtk.pyface.scene_editor import SceneEditor
import numpy
from itertools import chain, islice, tee

#from raypier.tracer import Optic, VTKOptic, normaliseVector, RaySegment,\
#             Traceable, NumEditor, dotprod, transformPoints, transformNormals

from raypier.bases import Optic, Traceable, NumEditor, RaypierObject
from raypier.core.cfaces import ElipticalPlaneFace, CircularFace
from raypier.core.ctracer import FaceList
from raypier.mirrors import BaseMirror
from raypier.core.cmaterials import PECMaterial, FullDielectricMaterial


class HollowRetroreflector(BaseMirror):
    abstract = False
    name = "Hollow Retroreflector"
    diameter = Float(25.4)

    imp_cyl = Instance(tvtk.Cylinder, (), transient=True)    
    scale = Instance(tvtk.Transform, (), transient=True)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('diameter', editor=NumEditor),
                        ),
                   )
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        sqrt2 = numpy.sqrt(2)
        x = sqrt2*numpy.cos(numpy.pi*2./3)
        y = sqrt2*numpy.sin(numpy.pi*2./3)
        m = self.material
        fl.faces = [ElipticalPlaneFace(owner=self, g_x=sqrt2, g_y=0, material=m),
                    ElipticalPlaneFace(owner=self, g_x=x, g_y=y, material=m),
                    ElipticalPlaneFace(owner=self, g_x=x, g_y=-y, material=m),
                    ]
        return fl
    
    @on_trait_change("diameter")
    def config_pipeline(self):
        r = self.diameter/2.
        cyl = self.imp_cyl
        cyl.radius = r
        s=self.scale
        s.identity()
        s.scale(r,r,r)
        
        self.update = True
    
    def _pipeline_default(self):
        sqrt2 = numpy.sqrt(2)
        x = -sqrt2*numpy.cos(numpy.pi*2./3)
        y = sqrt2*numpy.sin(numpy.pi*2./3)
        p1 = tvtk.PlaneSource(origin=(0,0,0), \
                              point1=(-sqrt2,0,1), point2=(x,y,1))
        p2 = tvtk.PlaneSource(origin=(0,0,0), \
                               point1=(x,y,1), point2=(x,-y,1))
        p3 = tvtk.PlaneSource(origin=(0,0,0), \
                               point1=(x,-y,1), point2=(-sqrt2,0,1))
        for p in (p1,p2,p3):
            p.set_resolution(32,32)
        append = tvtk.AppendPolyData()
        append.add_input_connection(p1.output_port)
        append.add_input_connection(p2.output_port)
        append.add_input_connection(p3.output_port)
        
        scale_f = tvtk.TransformFilter(input_connection=append.output_port, 
                                       transform=self.scale)
        
        trans = tvtk.Transform()
        trans.rotate_x(90.0)
        self.imp_cyl.transform = trans
        
        clip = tvtk.ClipPolyData(input_connection=scale_f.output_port, 
                                 clip_function = self.imp_cyl,
                                 inside_out=True)
                               
        transF2 = tvtk.TransformFilter(input_connection=clip.output_port, 
                                       transform=self.transform)
        self.config_pipeline()
        return transF2
    
    
class SolidRetroreflector(Optic, HollowRetroreflector):
    abstract = False
    yaml_tag = "!SolidRetroreflector"
    n_inside = 1.5
    name = "Solid Retroreflector"
    
    offset = 0.0
    
    thickness = Float(20.0, desc="distance from input face to apex")
    
    cyl = Instance(tvtk.CylinderSource, (), transient=True)
    cube = Instance(tvtk.CubeSource, (), transient=True)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('diameter', editor=NumEditor),
                       Item('thickness', editor=NumEditor),
                       Item('n_inside'),
                        ),
                   )
    
    def _material_default(self):
        print("Making full dielectric material")
        m = FullDielectricMaterial(n_inside = self.n_inside,
                                  n_outside = self.n_outside,
                                  reflection_threshold=0.5,
                                  transmission_threshold=0.5)
        return m
    
    def _faces_default(self):
        fl = super(SolidRetroreflector, self)._faces_default()
        m = self.material
        fl.faces.append(CircularFace(owner=self, material=m, 
                                     z_plane=self.thickness,
                                     invert_normal=True))
        print("adding circular face", m)
        return fl
    
    def _thickness_changed(self, new_t):
        face = self.faces.faces[-1]
        print("setting thickness", face, face.z_plane)
        face.z_plane = new_t
        thickness = new_t
        size = max(thickness, self.diameter)*2
        cube = self.cube
        cube.set_bounds(0,size,0,size,0,size)
        
        cyl=self.cyl
        cyl.height = thickness
        cyl.center = (0, thickness/2,0)
        self.update=True
        
    def _diameter_changed(self, d):
        self.cyl.radius = d/2.
        size = max(self.thickness, d)*2
        cube = self.cube
        cube.set_bounds(0,size,0,size,0,size)
        self.update=True
        
    def _pipeline_default(self):
        cyl = self.cyl
        cyl.resolution = 63 #use a prime no to avoid vertex degeneracy
        # Important to ensure the cylinder end doesn't coincide with the corner of the cube
        cyl.height = self.thickness-1
        cyl.radius = self.diameter/2.0
        cyl.center = (0, (self.thickness/2)-1,0)
        print("thickness", self.thickness)
        print("diameter", self.diameter)
        size = max(self.thickness, self.diameter)*2
        cube = self.cube
        cube.set_bounds(0,size,0,size,0,size)
        
        tf = tvtk.Transform()
        tf.post_multiply()
        tf.rotate_x(-45.0)
        tf.rotate_wxyz(35.26438968275, 0,0,1)
        
        tfilt = tvtk.TransformFilter(input_connection=cube.output_port)
        tfilt.transform=tf
        
        tri1 = tvtk.TriangleFilter(input_connection=cyl.output_port)
        tri2 = tvtk.TriangleFilter(input_connection=tfilt.output_port)
        tri1.update()
        tri2.update()
        
        intersect = tvtk.BooleanOperationPolyDataFilter()
        intersect.operation = "intersection"
        intersect.add_input_connection(0, tri1.output_port)
        intersect.add_input_connection(1, tri2.output_port)
        intersect.tolerance = 1e-8
        
        tf2=tvtk.Transform()
        tf2.rotate_x(90.0)
        tf2.rotate_y(60.0)
        orient = tvtk.TransformFilter(input_connection=intersect.output_port,
                                      transform=tf2)
        
        norm = tvtk.PolyDataNormals(input_connection=orient.output_port)
        transF = tvtk.TransformFilter(input_connection=norm.output_port, 
                                      transform=self.transform)
        self.config_pipeline()
        return transF
    


from traits.has_traits import HasTraits, \
        on_trait_change as _on_trait_change
        
from traits.trait_notifiers import TraitChangeNotifyWrapper

from collections import deque
from threading import currentThread
from functools import wraps


@wraps(_on_trait_change)
def on_trait_change ( name, post_init = False, 
                      dispatch="same", 
                      retrigger="all",
                      *names ):
    """ overrides the has_traits.on_trait_change decorator to
    add some extra metadata to the trait handler function to
    indicate the dispatch-method
    """
    def decorator( function ):
        function = _on_trait_change(name, post_init, *names)( function )
        function.dispatch = dispatch
        function.retrigger = retrigger
        return function
    return decorator


class QueuedTraitChangeNotifyWrapper(TraitChangeNotifyWrapper):
    def __init__(self, handler, owner, target=None):
        retrigger = getattr(handler, 'retrigger', 'all')
        self.dispatch = getattr(self, self._policy_map[retrigger])
        TraitChangeNotifyWrapper.__init__(self, handler, owner)
    
    def _dispatch_all(self, handler, *args):
        this_thd = currentThread()
        notification_queue = self.object().__notification_queue__
        this_queue = notification_queue[this_thd]
        this_queue.append((handler, args))
        
    def _dispatch_replace(self, handler, *args):
        this_thd = currentThread()
        notification_queue = self.object().__notification_queue__
        this_queue = notification_queue[this_thd]
        handlers = [h for h,a in this_queue]
        count = handlers.count(handler)
        for i in range(count):
            idx = [h for h,a in this_queue].index(handler)
            del this_queue[idx]
        this_queue.append((handler, args))
        
    _policy_map = {'all': '_dispatch_all',
                   'replace': '_dispatch_replace'}
            
            
def wrap_queue( func ):
    def wrapped( self, *args, **kwds ):
        s = super(HasQueue, self)
        q = self.__notification_queue__
        thd = currentThread()
        if thd not in q:
            q[thd] = this_q = deque()
            try:
                func(self, *args, **kwds)
                while True:
                    try:
                        handler, args = this_q.popleft()
                    except IndexError:
                        break
                    handler(*args)
            finally:
                del q[thd]
        else:
            func(self, *args, **kwds)
    return wrapped


class HasQueue(HasTraits):    
    """
    This subclass of HasTraits add a new method of trait notification
    dispatch: "queued". By default the scope of the queue is global across all
    subclasses of HasQueue but you can limit this by adding a __notification_queue__={}
    class attribute to specific subclasses, or even instances. 
    
    When any trait on the instance is set, the queue is processed after all
    synchronous trait handlers (dispatch="same") have completed.
    
    Use the on_trait_change decorator defined in this module to define the dispatch
    method (the has_trait.on_trait_change decorator doesn't allow a dispatch
    keyword).
    
    multiple firing of handlers will still occur (i.e. setting a handler to "queued"
    doesn't change the number of times it fires, only the ordering). If re-triggering is 
    not desired, this can easily be caught in the model logic.
    
    N.B. If *all* handlers are set to "queued" this is equivalent to a 
    breadth-first traversal of the dependancy tree (whereas all "same" gives
    a depth-first traversal).
    """
    __notification_queue__ = {}
    
    __init__ = wrap_queue( HasTraits.__init__ )
    
    trait_set = wrap_queue( HasTraits.trait_set )
        
    def _on_trait_change( self, handler, name = None, remove = False,
                                 dispatch = 'same', priority = False, target = None):
        dispatch = getattr(handler, "dispatch", dispatch)
        super(HasQueue, self)._on_trait_change( handler, name=name, 
                                                remove=remove,
                                                dispatch=dispatch, 
                                                priority=priority)
    
    def __setattr__(self, name, val):
        """is there a better way to intercept trait-assignment?"""
        s = super(HasQueue, self)
        if name in self.trait_names():
            q = self.__notification_queue__
            thd = currentThread()
            if thd not in q:
                q[thd] = this_q = deque()
                try:
                    s.__setattr__(name, val)
                    while True:
                        try:
                            handler, args = this_q.popleft()
                        except IndexError:
                            break
                        handler(*args)
                finally:
                    del q[thd]
            else:
                s.__setattr__(name, val)
        else:
            s.__setattr__(name, val)

HasQueue.set_trait_dispatch_handler("queued", QueuedTraitChangeNotifyWrapper)




if __name__=="__main__":
    from traits.api import Float, Bool
    class test(HasQueue):
        """
        A simple test object with a diamond-shape trait dependency
        tree. You can control the ordering of trait handler execution
        by changing the various dispatch methods to "same" or "queued".
        
        trait 'e' includes some logic to prevent re-triggering, as an example.
        """
        
        a = Float(0.0)
        b = Float(0.0)
        c = Float(0.0)
        d = Float(0.0)
        
        @on_trait_change("a", dispatch="same")
        def change_a(self, vnew):
            print("a changed to", vnew)
            self.c = vnew + 2
            self.b = vnew + 1
            
        @on_trait_change("b", dispatch="same")
        def change_b(self, vnew):
            print("b changed to", vnew)
            self.d = vnew + 3
            
        @on_trait_change("c", dispatch="same")
        def change_c(self, vnew):
            print("c changed to", vnew)
            self.d = vnew + 4
            
        @on_trait_change("d", dispatch="queued", retrigger="replace")
        def change_d(self, vnew):
            print("d changed to", vnew)
            self.e = vnew + 5
            
    t = test()
    t.a = 1.0



import unittest

import pyximport
pyximport.install()

from raypier.sources import SingleRaySource
from raypier.diffraction_gratings import RectangularGrating
from raypier.tracer import RayTraceModel
from raypier.core.ctracer import cross

import numpy


class TestGratingEquation(unittest.TestCase):
    def test_grating_equation1(self):
        grating = RectangularGrating(centre=(0,0,0),
                                     direction=(0,1,0),
                                     order=-1,
                                     lines_per_mm=1400)
        source = SingleRaySource(origin=(-10,10,0),
                                 direction=(1,-1,0),
                                 wavelength=0.8,
                                 )
        model = RayTraceModel(sources=[source,],
                              optics=[grating,])
        model.trace_all()
        traced_rays = list(source.traced_rays)
        self.assertEqual(len(traced_rays),2)
        
        out = traced_rays[-1]
        in_ray = traced_rays[0][0]
        out_ray = out[0]
        
        
        sin_in = -cross(in_ray.direction, grating.direction)[2]
        sin_out = cross(out_ray.direction, grating.direction)[2]
        
        m = grating.order
        D=1000./grating.lines_per_mm #in microns
        wl=0.8 #wavelength in microns
        LHS = m*wl-(D*sin_in)
        RHS = D*sin_out
        self.assertAlmostEqual(abs(LHS), abs(RHS))
        
    def test_grating_phase(self):
        grating = RectangularGrating(centre=(0.3e-3,0.3e-3,0.4e-3),
                                     direction=(0,1,0),
                                     order=-1,
                                     lines_per_mm=1400)
        source = SingleRaySource(origin=(-10,10,0),
                                 direction=(1,-1,0),
                                 wavelength=0.8,
                                 )
        model = RayTraceModel(sources=[source,],
                              optics=[grating,])
        model.trace_all()
        traced_rays = list(source.traced_rays)
        self.assertEqual(len(traced_rays),2)
        
        out = traced_rays[-1]
        in_ray = traced_rays[0][0]
        out_ray = out[0]
        
        print(out_ray.phase)

        


from traits.api import HasTraits, Array, Float, Complex,\
            Property, List, Instance, on_trait_change, Range, Any,\
            Tuple, Event, cached_property, Set, Int, Trait, Bool

from traitsui.api import View, Item, ListEditor, VSplit,\
            RangeEditor, ScrubberEditor, HSplit, VGroup, Heading
from tvtk.api import tvtk
from tvtk.pyface.scene_model import SceneModel
from tvtk.pyface.scene_editor import SceneEditor

from raypier.core.cmaterials import OpaqueMaterial

from raypier.bases import Traceable
from raypier.prisms import Extrusion
from raypier.splines import Extruded_interpolant
from raypier.core.cfaces import PolygonFace, ExtrudedPlanarFace

from raypier.core.ctracer import FaceList
import numpy as np 

class LDLF(Extrusion):
    name = "Linear Dialectric Light Funnle"
    slat_width= Float #width of slats
    ap_width = Float #width of exit apperture
    slant = Float #angle of slanted sides
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('trace_ends'),
                       Item('n_inside'),
                       Item('length'),
                       Item('slat_width'),
                       Item('ap_width'),
                       Item('slant')
                       )
                       )

    @on_trait_change("slat_width, ap_width, slant")
    def config_profile(self):
        theta = self.slant*np.pi/180.
        l = self.slat_width
        h = l*np.sin(theta)
        dis = l*np.cos(theta)
        w = self.ap_width/2
        self.profile = [(-w,0),
                        (w,0),
                        (w+dis,h),
                        (-w-dis,h)]

class CylindricalLDLF(Extruded_interpolant):
    #subclassing the Bezier extrusion to build a multipart optical element.
    name = "Cylindrical Linear Dialectric Light Funnle"
    entry_width = Float
    radius = Float
    slant = Float
    depth = Float
    aperture_material = OpaqueMaterial()


    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = self.make_faces()
        return fl

    def make_faces(self):
        radius = self.radius
        opening = .5*self.entry_width
        d=self.depth
        slant = self.slant
        z1 = self.z_height_1
        z2 = self.z_height_2
        m=self.material

        if radius**2 < opening**2:
            print("error: radius is too small (<.5 width)")
            return

        if d/np.tan(np.radians(slant)) > opening:
            print("The Cyclindrical LDLF is too deep, slanted sides intersect.")
            return

        sign = 1
        if radius < 0:
            sign = -1
        
        #lens_profile
        x = np.linspace(-opening,opening)
        displace = np.sqrt(radius**2-opening**2)
        ys = sign*(np.sqrt(radius**2-x**2)-displace)
        self.profile = [x,ys]
        self.trace_ends=True
        self.trace_top = False
        sides = Extruded_interpolant.make_faces(self)    #list of faces

        #one flat, slanty side
        x1 = opening
        y1 = 0
        x2 = opening-(d/np.tan(np.radians(slant)))
        y2 = -d

        sides.append(ExtrudedPlanarFace(owner=self, z1=z1, z2=z2, x1=x1, y1=y1, 
                    x2=x2, y2=y2, material=m))

        #the other, symmetrical one
        sides.append(ExtrudedPlanarFace(owner=self, z1=z1, z2=z2, x1=-x1, y1=y1, 
                    x2=-x2, y2=y2, material=m,invert_normal=True))

        #the target
        sides.append(ExtrudedPlanarFace(owner=self, z1=z1, z2=z2, x1=x2, y1=y2, 
                    x2=-x2, y2=y2, material=self.aperture_material))
        
        
        if self.trace_ends:
            #warning, just polygon approximations.  not water tight.
            profile = self.get_real_profile()
            profile[0] = np.append(profile[0], [x1,x2,-x2,-x1],1)
            profile[1] = np.append(profile[1], [y1,y2,y2,y1],1)
            base = PolygonFace(owner=self, z_plane=z1,
                        xy_points=profile, material=m)
            top = PolygonFace(owner=self, z_plane=z2, material=m,
                        xy_points=profile, invert_normal=True)
            sides.extend([base, top])

        return sides

    def _pipeline_default(self):
        #pretty much just copied this from extruded_interpolant, and added three profile points
        source = self.data_source
        def execute():

            opening = .5*self.entry_width
            d=self.depth
            slant = self.slant

            x1 = opening
            y1 = 0
            x2 = opening-(d/np.tan(np.radians(slant)))
            y2 = -d

            profile = self.get_real_profile()
            profile[0] = np.append(profile[0], [x2,-x2],1)
            profile[1] = np.append(profile[1], [y2,y2],1)
            xy = np.column_stack((profile[:][0],profile[:][1]))
            z = np.ones(xy.shape[0]) * self.z_height_1
            points = np.column_stack((xy,z))
            
            cells = [list(range(len(z))),]
            
            output = source.poly_data_output
            output.points = points
            output.polys = cells
        source.set_execute_method(execute)
        
        self.extrude.scale_factor = self.z_height_2 - self.z_height_1  #mm, put here because it wasn;t being initialized
        if self.trace_ends:
            print("drew ends")
            self.extrude.capping = True 
        extrude = self.extrude
        extrude.input = source.output
        
        t = self.transform
        transf = tvtk.TransformFilter(input_connection=extrude.output_port, 
                                      transform=t)
        return transf






"""
Traited wrappers for various dispersion-curve definitions
"""


from .dispersion import NondispersiveCurve, NamedDispersionCurve

from traits.api import HasStrictTraits, Instance, Float, Str, Enum, Property, cached_property, Bool

from traitsui.api import View, VGroup, Item

from .editors import NumEditor


GlassNames = sorted(NamedDispersionCurve.get_glass_names())



class OpticalMaterial(HasStrictTraits):
    name = Str("Part A")
    from_database = Bool(True)
    refractive_index = Float(1.0)
    glass_name=Enum(GlassNames)
    absorption=Float(0.0)
    
    dispersion_curve = Property(depends_on="glass_name, absorption, refractive_index, from_database")
    
    traits_view = View(VGroup(
            Item("from_database"),
            Item("glass_name", style="simple", enabled_when="from_database"),
            Item("refractive_index", editor=NumEditor, enabled_when="not from_database"),
            Item("absorption", editor=NumEditor),
        ))
    
    def __init__(self, *args, **kwds):
        if "from_database" not in kwds:
            if ("refractive_index" in kwds) and ("glass_name" not in kwds):
                kwds['from_database'] = False
        return super().__init__(*args, **kwds)
    
    @cached_property
    def _get_dispersion_curve(self):
        if self.from_database:
            return NamedDispersionCurve(self.glass_name, absorption=self.absorption)
        else:
            return NondispersiveCurve(refractive_index=self.refractive_index,
                                  absorption=self.absorption)
    
    
class AirMaterial(OpticalMaterial):
    from_database = False
    name = "Air"
    
    
air = AirMaterial()




from raypier.results import TargetResult, evaluate_phase
from raypier.dispersion import FusedSilica

from traits.api import Float, Instance, Bool
from traitsui.api import View, Item, VGroup, HGroup
from chaco.api import Plot, ArrayPlotData
from enable.api import ComponentEditor
from chaco.tools.api import PanTool, ZoomTool

import numpy as np
from numpy import fft
from numpy.core._umath_tests import euclidean_pdist
from traits.has_traits import on_trait_change
from posix import pwrite



class ChirpResult(TargetResult):
    name = "Pulse Intensity Envelope"
    plot = Instance(Plot)
    
    #: The input pulse width, in picoseconds
    pulse_width = Float(0.2) #in picoseconds
    
    #: The length of additional fibre (fused silica) in the optical path, in metres
    glass_path = Float(0.0) #glass path in metres
    
    #: Display the autocorrelation function
    ACF = Bool(False)
    
    #: Force the output pulse peak to be at the centre of the array
    centre = Bool(True)
    
    #: Calculated FWHM for the output pulse width
    fwhm = Float(0.0)
    
    #: Centre wavelength, in microns.
    wavelength = Float(0.780) #in microns
    
    _fs = Instance(FusedSilica, ())
    
    _pd = Instance(ArrayPlotData, ())
    
    traits_view = View(VGroup(HGroup(Item("pulse_width", tooltip="input pulse width (picoseconds)."), 
                                     Item("glass_path", tooltip="Fused silica glass path (metres)")),
                              HGroup(Item("ACF", tooltip="Autocorrelation Function"), Item("centre"),
                              Item("fwhm", label="FWHM", style="readonly", 
                                   tooltip="femtoseconds"))
                              ),
                       Item("plot", editor=ComponentEditor(), show_label=False))
    
    def _plot_default(self):
        x = np.linspace(-14, 14, 100)
        y = np.sin(x) * x**3
        plotdata = self._pd
        plotdata.set_data("x", x)
        plotdata.set_data("y", y)

        plot = Plot(plotdata)
        plot.plot(("x", "y"), type="line", color="blue")
        #plot.title = "sin(x) * x^3"
        plot.tools.append(PanTool(plot, drag_button="right"))
        plot.overlays.append(ZoomTool(plot))
        self._calc_result(redraw=False)
        return plot
    
    @on_trait_change("glass_path, ACF, pulse_width")
    def do_update(self):
        self.update()
    
    def _calc_result(self, redraw=True):
        all_wavelengths = np.asarray(self.source.wavelength_list)
        traced_rays = self.source.traced_rays
        target_face = self.target
        glass_length = self.glass_path
        glass_dispersion = self._fs
        f, phase = evaluate_phase(all_wavelengths, traced_rays, target_face, glass_length, glass_dispersion)
        
        c = 2.99792458e8 * 1e-9 #convert to mm/ps
        hsize = len(phase)//2
        dphi = np.diff(phase)[hsize-hsize//5:hsize + hsize//5].mean()
        phase -= np.arange(len(phase))*dphi
        phase -= phase[hsize]
        
        f0 = c/(self.wavelength*0.001) #in THz
        dt = 0.2*(1/f0)/3.0 #in ps
        size = int(20.0* self.pulse_width/dt)
        #print("SIZE:", size, "F0", f0)
        t = np.arange(size)*dt - 10*self.pulse_width
        omega0 = 2*np.pi*f0
        pulse_in = np.exp(-(t**2)/((self.pulse_width/2)**2))*np.exp(1.0j*omega0*t)
        
        spec = fft.fft(pulse_in)
        f_max = 1./(2*dt)
        f2 = np.roll(np.linspace(-f_max,f_max,len(spec)), len(spec)//2)
        
        phase2 = np.interp(f2, f, phase, 0.0, 0.0) + np.interp(-f2, f, phase, 0.0, 0.0)
        
        bad = np.logical_or(np.abs(f2)<f.min(), np.abs(f2)>f.max())
        spec *= np.exp(1.0j*phase2/1.0)
        spec[bad] = 0.0
        
        pulse_out = fft.ifft(spec)
        
        pwr = pulse_out.real**2 + pulse_out.imag**2
        
        if self.ACF:
            psp = fft.rfft(pwr)
            pwr = fft.irfft(psp * psp.conjugate())
            pwr = np.roll(pwr, len(pwr)//2)
            
        if self.centre:
            imax = np.argmax(pwr)
            middle = len(pwr)//2
            pwr = np.roll(pwr, middle-imax)
        
        mask = np.arange(len(pwr))[(pwr > (pwr.max()/2))]
        fwhm = t[mask.max()+1] - t[mask.min()]
        self.fwhm = fwhm*1000 #convert to fs
        
        self._pd.set_data('x', t)
        self._pd.set_data('y', pwr)
        if redraw: #prevent a recursive loop.
            self.plot.request_redraw()
        


from raypier.mirrors import PECMirror
from raypier.sources import ConfocalRaySource
from raypier.tracer import RayTraceModel

m1 = PECMirror(name="M1",
                centre=(0,-45, 0),
                direction=(0.,1,-1), #(0,y_dir,-small_height),
                diameter=40.,
                thickness=4.)

source = ConfocalRaySource(focus=(0,0,0),
                            direction=(0,1,0),
                            working_dist = 100.,
                            number=20,
                            detail_resolution=5,
                            theta=15.00, #fringe ray is at 15.22 deg
                            scale_factor=0.2)

model = RayTraceModel(sources=[source],
                      optics=[m1])

model.configure_traits()
#model.update = True
#print source.traced_rays






from .shapes import CircleShape, RectangleShape, PolygonShape, HexagonShape

from .general_optic import GeneralLens

from .faces import PlanarFace, SphericalFace, CylindericalFace, AsphericFace, ConicFace, \
        AxiconFace, DistortionFace
        
from .lenses import PlanoConvexLens, PlanoConicLens, AsphericLens, BiConicLens
        
from .materials import OpticalMaterial, air
        
from .sources import ConfocalRaySource, ConfocalRayFieldSource, ParallelRaySource,\
        GaussianBeamRaySource, SingleRaySource, HexagonalRayFieldSource, AdHocSource,\
        BroadbandRaySource
        
from .gausslet_sources import CollimatedGaussletSource

from raypier.decompositions import AngleDecompositionPlane, PositionDecompositionPlane

from .tracer import RayTraceModel

from .fields import EFieldPlane

from .decompositions import AngleDecompositionPlane, PositionDecompositionPlane

from .probes import RayCapturePlane, GaussletCapturePlane

from .apertures import CircularAperture, RectangularAperture

from .intensity_surface import IntensitySurface

from .beamsplitters import UnpolarisingBeamsplitterCube, PolarisingBeamsplitterCube

from .constraints import Constraint

from .core.ctracer import ray_dtype, gausslet_dtype, RayCollection, GaussletCollection
from .core.fields import eval_Efield_from_gausslets, eval_Efield_from_rays





from traits.api import Float, on_trait_change, Range, Instance

from traitsui.api import View, VGroup, Item

from raypier.bases import Optic, Traceable, NumEditor
from raypier.core.cfaces import CircularFace
from raypier.core.ctracer import FaceList
from raypier.core.cmaterials import WaveplateMaterial
            
import numpy
from tvtk.api import tvtk


class Waveplate(Optic):
    abstract=False
    name = "Waveplate"
    retardance = Range(0.0, 1.0, value=0.25)
    diameter = Float(10.0)
    offset = Float(0.0)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('retardance', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       Item('offset', editor=NumEditor),
                        ),
                   )
    
    vtk_disk = Instance(tvtk.DiskSource, (),
                            dict(circumferential_resolution=64,
                                 inner_radius=0), transient=True)
    cyl_trans = Instance(tvtk.Transform, (), transient=True)
    
    line = Instance(tvtk.LineSource, (), transient=True)
    
    def _retardance_changed(self, val):
        self.material.retardance = val
    
    def _faces_default(self):
        m = self.material
        fl = FaceList(owner=self)
        fl.faces = [CircularFace(owner=self, z_plane=0, material=m)]
        return fl
    
    def _material_default(self):
        m = WaveplateMaterial(retardance=self.retardance,
                                fast_axis = self.x_axis)
        return m
    
    @on_trait_change("_orientation, rotation")
    def update_material_fast_axis(self):
        self.material.fast_axis = self.x_axis
        
    @on_trait_change("diameter, offset")
    def config_pipeline(self):
        rad = self.diameter/2.
        cyl = self.vtk_disk
        cyl.outer_radius = rad
        
        t = self.cyl_trans
        t.identity()
        t.translate(self.offset,0,0)
        #t.rotate_x(90.)
        
        line = self.line
        line.point1 = [-rad, 0, 0]
        line.point2 = [rad, 0, 0]
        
        self.update = True
        
    def _pipeline_default(self):
        cyl = self.vtk_disk
        norms = tvtk.PolyDataNormals(input_connection=cyl.output_port)
        
        tube = tvtk.TubeFilter(input_connection=self.line.output_port)
        
        append = tvtk.AppendPolyData()
        append.add_input_connection(norms.output_port)
        append.add_input_connection(tube.output_port)
        
        transF1 = tvtk.TransformFilter(input_connection=append.output_port, 
                                       transform=self.cyl_trans)
        
        transF2 = tvtk.TransformFilter(input_connection=transF1.output_port, 
                                       transform=self.transform)
        self.config_pipeline()
        return transF2



from .ctracer import trace_segment, trace_gausslet, RayCollection

from itertools import chain
import numpy


def trace_rays(input_rays, face_lists, recursion_limit=100, max_length=100.0):
    """
    Core ray-tracing routine. Takes either a RayCollection or GaussletCollection
    and traces the rays non-sequentially through the given list of FaceList objects.
    
    The input_rays should already have a consistent wavelengths property set.
    
    returns - (traced_rays, all_faces)
            where traced_rays is a list of RayCollection or GaussletCollections 
            representing the sequence of ray generations. The 'all_faces' list
            is a list of cfaces.Face objects. The end_face_idx member of each
            traced ray indexes into this list to give the face where it terminates.
    """
    input_rays.reset_length(max_length)
    traced_rays = []
    trace_func = trace_segment if isinstance(input_rays, RayCollection) else trace_gausslet
    count = 0
    wavelengths = numpy.asarray(input_rays.wavelengths)
    
    all_faces = list(chain(*(fs.faces for fs in face_lists)))
    for i, f in enumerate(all_faces):
        f.idx = i
        f.count = 0 #reset intersection count
        f.update() #This was a bad idea.
        f.material.wavelengths = wavelengths
        f.max_length = max_length
    
    face_sets = list(face_lists)
    decomp_faces = [f for f in all_faces if f.material.is_decomp_material()]
    
    rays = input_rays
    while rays.n_rays>0 and count<recursion_limit:
        traced_rays.append(rays)
        rays = trace_func(rays, face_sets, all_faces, 
                                     max_length=max_length,
                                     decomp_faces=decomp_faces)
        count += 1
    
    return traced_rays, all_faces 





from traits.api import Float, Instance, \
        List, observe, Tuple, Str, ComparisonMode

from traitsui.api import View, Item, VGroup, Group, ListEditor

from tvtk.api import tvtk
import numpy
from operator import attrgetter

from raypier.bases import NumEditor, Traceable

from raypier.core.cmaterials import CoatedDispersiveMaterial, PECMaterial
from raypier.shapes import BaseShape
from raypier.lenses import BaseLens
from raypier.core.ctracer import FaceList
from . import faces
from .materials import OpticalMaterial, air


alist_editor = ListEditor(use_notebook=True,
                           deletable=False,
                           selected='selected',
                           export='DockWindowShell',
                           page_name='.name')


class GeneralLens(BaseLens):
    #: The name of the object in the model-tree
    name = Str("General Lens")
    
    #: An instance of a :py:class:`raypier.shapes.BaseShape` subclass, defining
    #: the 2D outline of the optic.
    shape = Instance(BaseShape)
    
    #: A list of :py:class:`raypier.faces.BaseFace` objects which defines the
    #: geometry of the surfaces comprising this optic.
    surfaces = List(faces.BaseFace, comparison_mode=ComparisonMode.identity)
    
    #: A list of :py:class:`raypier.materials.OpticalMaterial` objects describing 
    #: the materials properties of the dielectrics between the previously given
    #: surfaces. For a list of N surfaces, you should provide (N-1) materials.
    materials = List(OpticalMaterial, comparison_mode=ComparisonMode.identity)
    
    _grid = Instance(tvtk.PlaneSource, ())
    _clip = Instance(tvtk.ClipPolyData, ())
    
    #: A :py:class:`raypier.materials.OpticalMaterial` object to define the 
    #: characteristics of a single-layer dielectric coating applied to the outer
    #: surfaces of the object.
    coating_material = Instance(OpticalMaterial, ())
    
    #: The thickness, in microns, of the coating on the outer surfaces.
    coating_thickness = Float(0.25, desc="Thickness of the AR coating, in microns")
    
    
    grid_resolution = Tuple((200,200))
    glass_colour = Tuple((204,204,255,102))
    mirror_colour = Tuple((104,64,104,255))
    
    
    
    traits_view = View(Group(
                Traceable.uigroup,
                Group(
                    Item("shape", show_label=False, style="custom"),
                    label="Outline", dock="tab"
                    ),
                Group(
                    Item("surfaces", style="custom",
                         editor=alist_editor, 
                         show_label=False),
                    label="Faces", dock="tab"
                    ),
                Group(
                    VGroup(
                    Item("materials", style="custom", editor=alist_editor, show_label=False),
                    ),
                    Item("coating_material", style="custom"),
                    Item("coating_thickness", editor=NumEditor),
                    label="Materials", dock="tab"
                    ),
                layout="tabbed"
            )
        )
    
    def _actors_default(self):
        pipeline = self.pipeline
        
        map = tvtk.PolyDataMapper(input_connection=pipeline.output_port,
                                  scalar_mode="use_cell_data",
                                  color_mode="direct_scalars")
        map.scalar_visibility = True
        act = tvtk.Actor(mapper=map)
        actors = tvtk.ActorCollection()
        actors.append(act)
        return actors
    
    def _pipeline_default(self):
        nx,ny = self.grid_resolution
        xmin,xmax,ymin,ymax,zmin,zmax = self.grid_extent
        
        shape = self.shape
        func = shape.impl_func
        
        _grids = []
        
        append = tvtk.AppendPolyData()
        
        grid = self._grid
        grid.origin = (xmin,ymin,0)
        grid.point1=(xmin,ymax,0)
        grid.point2=(xmax,ymin,0)
        grid.x_resolution=nx
        grid.y_resolution=ny
    
        clip = self._clip
        clip.inside_out=True
        clip.input_connection = grid.output_port
        clip.clip_function = func
        
        bounds = tvtk.FeatureEdges(input_connection=clip.output_port)
        bounds.extract_all_edge_types_off()
        bounds.boundary_edges = True
        bounds.coloring = False
        
        for face in self.surfaces:
            attrb = tvtk.ProgrammableAttributeDataFilter(input_connection=clip.output_port)
            
            ###Beware, references to outer scope change in the loop. Capture state using kwd-args.
            def execute( *args , _face=face, _attrb=attrb):
                in_data = _attrb.get_input_data_object(0,0)
                points = in_data.points.to_array()
                z = _face.cface.eval_z_points(points.astype('d'))
                out = _attrb.get_output_data_object(0)
                out.point_data.scalars=z
                ncells = in_data.polys.number_of_cells
                colors = numpy.zeros((ncells,4), numpy.uint8)
                if _face.mirror:
                    colors[:] = self.mirror_colour
                else:
                    colors[:] = self.glass_colour
                out.cell_data.scalars = colors

            attrb.set_execute_method(execute)
            
            warp = tvtk.WarpScalar(input_connection=attrb.output_port,scale_factor=1.0)
            warp.normal=(0,0,1)
            warp.use_normal=True
            
            append.add_input_connection(warp.output_port)
                
        skirt = tvtk.ProgrammableFilter(input_connection=bounds.output_port)
        def calc_skirt(_skirt=skirt):
            in_data = _skirt.get_input_data_object(0,0)
            out = _skirt.get_output_data_object(0)
            points = in_data.points.to_array().astype('d')
            z_top = self.surfaces[0].cface.eval_z_points(points)
            z_bottom = self.surfaces[-1].cface.eval_z_points(points)
            
            size = points.shape[0]
            points_out = numpy.vstack([points,points])
            points_out[:size,2] = z_top
            points_out[size:,2] = z_bottom
            
            cell_data = in_data.lines.to_array().reshape(-1,3)
            ncells = cell_data.shape[0]
            cells_out = numpy.empty((ncells, 5))
            cells_out[:,0] = 4
            cells_out[:,1] = cell_data[:,1]
            cells_out[:,2] = cell_data[:,2]
            cells_out[:,3] = cell_data[:,2] + size
            cells_out[:,4] = cell_data[:,1] + size    
            out.points = points_out.reshape(-1,3)
            quads = tvtk.CellArray()
            quads.set_cells(5, cells_out)
            out.polys = quads
            out.lines = None

            colors = numpy.zeros((ncells*3,4), numpy.uint8)
            colors[:] = self.glass_colour
            out.cell_data.scalars = colors
            
        skirt.set_execute_method(calc_skirt)
        
        clean = tvtk.CleanPolyData(input_connection=skirt.output_port)
        
        append.add_input_connection(skirt.output_port)
        
        norms = tvtk.PolyDataNormals(input_connection=append.output_port)
        
        transF = tvtk.TransformFilter(input_connection=norms.output_port, 
                                      transform=self.transform)
        return transF
            
            
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.config_cfaces()
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        return fl
    
    def config_cfaces(self):
        surfaces = sorted(self.surfaces, key=attrgetter("z_height"))
        cfaces = []
        mats = [air,] + self.materials + [air,]
        if len(mats) < len(surfaces) + 1:
            mats += [air,] * (len(surfaces)-len(mats) + 1)
        ###FIXME: need to iterate in order of z_height to get the
        ###the materials on the correct sides.
        for i, face in enumerate(surfaces):
            if face.trace:
                cface = face.cface
                cface.owner = self
                cface.shape = self.shape.cshape
                if face.mirror:
                    cface.material = PECMaterial()
                else:
                    mat = CoatedDispersiveMaterial()
                    mat.dispersion_inside = mats[i].dispersion_curve
                    mat.dispersion_outside = mats[i+1].dispersion_curve
                    if i in {0, len(surfaces)-1}:
                        mat.dispersion_coating = self.coating_material.dispersion_curve
                        mat.coating_thickness = self.coating_thickness
                    else:
                        mat.coating_thickness = 0.0
                    cface.material = mat
                cfaces.append(face.cface)
        self.faces.faces = cfaces
    
    @observe("materials.items.dispersion_curve, coating_material.dispersion_curve, coating_thickness")
    def on_material_change(self, evt):
        self.config_cfaces()
        self.update = True
        
    @observe("surfaces.items.updated, shape.updated")
    def on_surfaces_changed(self, evt):
        self.config_cfaces()
        print("Face changed", evt)
        self.on_bounds_change(None)
        self.update = True

    @observe("shape.impl_func, surfaces")
    def on_impl_func_change(self, evt):
        self._clip.clip_function = self.shape.impl_func
        self.render=True
    
    @observe("shape.bounds, surfaces")
    def on_bounds_change(self, evt):
        nx, ny = self.grid_resolution
        xmin, xmax, ymin, ymax = self.shape.bounds
        zmin, zmax = self.eval_z_extent(xmin, xmax, ymin, ymax)
        self.grid_extent = (xmin, xmax, ymin, ymax, zmin, zmax)
        g = self._grid
        g.origin=(xmin,ymin,0)
        g.point1=(xmin,ymax,0)
        g.point2=(xmax,ymin,0)
        g.x_resolution=nx
        g.y_resolution=ny
        g.modified()
    
    def eval_z_extent(self, xmin, xmax, ymin, ymax):
        nx, ny = self.grid_resolution
        try:
            top_face = self.surfaces[0]
            bottom_face = self.surfaces[-1]
        except IndexError:
            return 0,0
        x = numpy.linspace(xmin, xmax, nx)
        y = numpy.linspace(ymin, ymax, ny)
        tzmin, tzmax = top_face.cface.eval_z_extent(x,y)
        bzmin, bzmax = bottom_face.cface.eval_z_extent(x,y)
        return ( min(tzmin,bzmin), max(tzmax, bzmax) )
    
    


"""
Functions relating to the creation and manipulations of gausslets
"""

import numpy
import time

from .core.utils import normaliseVector
from .core.ctracer import FaceList
from .bases import Traceable
from .core.cmaterials import ResampleGaussletMaterial
from .core.cfaces import CircularFace
from .core.fields import eval_Efield_from_gausslets
from .core.gausslets import make_hexagonal_grid, decompose_angle

from traits.api import Range, Float, Array, Property, Instance, observe, Str
from traitsui.api import View, Group, Item

from tvtk.api import tvtk
from raypier.editors import NumEditor, IntEditor
from raypier.core.gausslets import decompose_position




class BaseDecompositionPlane(Traceable):
    ### Sets the geometry of the intersection face
    diameter = Float(25.0)
    offset = Float(0.0)
    
    material = Instance(ResampleGaussletMaterial)
    
    def _material_default(self):
        m = ResampleGaussletMaterial(eval_func=self.evaluate_decomposed_rays)
        return m
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces =  [CircularFace(owner=self, z_plane=0.0, material=self.material)]
        return fl
        
    def _pipeline_default(self):
        radius = self.radius
        circle = tvtk.ArcSource(use_normal_and_angle=True,
                                center=[0.0,0.0,0.0],
                                polar_vector=[0.0, radius, 0.0],
                                normal=[0.0,0.0,1.0],
                                angle=360.0,
                                resolution=100
                                )
        
        trans = tvtk.TransformFilter(input_connection=circle.output_port,
                                     transform=self.transform)
        return trans
    
    def evaluate_decomposed_rays(self, input_rays):
        raise NotImplementedError()


class PositionDecompositionPlane(BaseDecompositionPlane):
    abstract=False
    name = Str("Position Decomposition")
    
    radius = Property()
    
    ### Wavefront curvature estimate, as a radius from the focus to decomposition plane centre
    curvature = Float(0.0)
    
    resolution = Float(10.0)
    
    blending = Float(1.2)
    
    ### Keep the sample points and sampled fields around 
    ### for debugging purposes
    _sample_points = Array()
    _unwrapped_phase = Array()
    _E_field = Array()
    
    traits_view = View(Group(
                    Traceable.uigroup,
                    Item("radius", editor=NumEditor),
                    Item("resolution", editor=NumEditor),
                    Item("curvature", editor=NumEditor),
                    Item("blending", editor=NumEditor),
                    ))
    
    def _get_radius(self):
        return self.diameter/2.
    
    def _set_radius(self, val):
        self.diamter = val*2.0
        
    @observe("radius, curvature, resolution, blending")
    def _on_changed(self, evt):
        self.update=True
    
    def evaluate_decomposed_rays(self, input_rays):
        start_time = time.monotonic()
        radius = self.diameter/2.
        curvature = self.curvature
        resolution = self.resolution
        blending = self.blending
        
        origin = numpy.asarray(self.centre)
        direction = normaliseVector(numpy.asarray(self.direction))
        axis1 = normaliseVector(numpy.asarray(self.x_axis))
        
        if curvature==0.0:
            curvature=None
        
        gausslets, E_field, uphase = decompose_position(input_rays, origin, direction, axis1, radius, \
                                                        resolution, curvature=curvature, blending=blending)
        
        self._unwrapped_phase = uphase
        self._E_field = E_field
        end_time = time.monotonic()
        print(f"Decomposition in {end_time-start_time} seconds")
        return gausslets


class AngleDecompositionPlane(BaseDecompositionPlane):
    abstract=False
    name = Str("Angle Decomposition")
    
    sample_spacing = Float(1.0) #in microns
    
    width = Range(0,512,64)
    height = Range(0,512,64)
    
    _mask = Instance(numpy.ndarray)
    mask = Property()
    
    ### Maximum ray ange in degrees
    max_angle = Float(30.0)
    
    _E_field = Array() #store the E_field at the aperture for debugging purposes
    _k_field = Array()
    
    traits_view = View(Group(
                    Traceable.uigroup,
                    Item("sample_spacing", editor=NumEditor),
                    Item("width", editor=IntEditor),
                    Item("height", editor=IntEditor),
                    Item("max_angle", editor=NumEditor)
                    ))
    
    def evaluate_decomposed_rays(self, input_rays):
        origin = numpy.asarray(self.centre)
        direction = numpy.asarray(self.direction)
        axis1 = numpy.asarray(self.x_axis)
        axis2 = numpy.cross(axis1, direction)
        spacing = self.sample_spacing
        
        wavelengths = input_rays.wavelengths
        
        x = numpy.arange(self.width)*spacing/1000.0
        x -= x.mean()
        y = numpy.arange(self.height)*spacing/1000.0
        y -= y.mean()
        
        points = x[:,None,None]*axis1[None,None,:] + y[None,:,None]*axis2[None,None,:]
        points += origin[None,None,:]
        
        flat_points = points.reshape(-1,3)
        E_field = eval_Efield_from_gausslets(input_rays, flat_points).reshape(*points.shape)
        self._E_field = E_field
        mask = self._mask
        if mask is not None:
            E_field *= mask[:,:,None]
        max_angle = self.max_angle
        
        pwr = (E_field.real**2 + E_field.imag**2).sum(axis=-1)
        imax = pwr.argmax()
        E_max = E_field.reshape(-1,3)[imax,:]
        p_max = flat_points[imax,:]
        
        new_rays, debug_k = decompose_angle(origin, direction, axis1, E_field, spacing, 
                                   max_angle, wavelengths[0], oversample=4, E_max=E_max, pos_max=p_max)
        print("Decomposed rays count:", len(new_rays))
        self._k_field = debug_k
        return new_rays
    
    def _vtkproperty_default(self):
        return tvtk.Property(color=(0.1,0.1,0.1) )
    
    def _get_mask(self):
        return self._mask
    
    def _set_mask(self, obj):
        w,h = obj.shape
        self._mask = numpy.clip(obj.astype('d'), 0.0, 1.0)
        self.width = w
        self.height = h
        
    def set_circular_mask(self, diameter):
        w = self.width
        h = self.height
        x = numpy.arange(w)*self.sample_spacing
        x -= x.mean()
        y = numpy.arange(h)*self.sample_spacing
        y -= y.mean()
        mask = numpy.zeros((w,h))
        X,Y = numpy.meshgrid(x,y)
        inside = (X**2) + (Y**2) <= ((diameter/2)**2)
        mask[inside] = 1.0
        self.mask = mask
        self.update = True
        
    def set_rectangular_mask(self, width, height):
        w = self.width
        h = self.height
        x = numpy.arange(w)*self.sample_spacing
        x -= x.mean()
        y = numpy.arange(h)*self.sample_spacing
        y -= y.mean()
        X,Y = numpy.meshgrid(x,y)
        inside = numpy.maximum(numpy.abs(X)-(width/2), numpy.abs(Y)-(height/2)) <= 0.0
        mask = numpy.zeros((w,h))
        mask[inside] = 1.0
        self.mask = mask
        self.update = True
        
    @observe("width, height")
    def _chk_dims(self, evt):
        mask = self._mask
        if mask is None:
            return
        if (self.width, self.height) != mask.shape:
            raise ValueError("Width and height must match shape of mask.")
    
    





import numpy

from numpy import fft
from scipy.interpolate import RectBivariateSpline
from scipy.sparse.linalg import lsqr, lsmr
from scipy.sparse import coo_matrix

from .ctracer import ray_dtype, GaussletCollection
from .utils import normaliseVector
from .fields import eval_Efield_from_gausslets, EFieldSummation
from .unwrap2d import unwrap2d
from .cfields import calc_mode_curvature, build_interaction_matrix, apply_mode_curvature


root_pi = numpy.sqrt(numpy.pi)

def next_power_of_two(n):
    return int(2**numpy.ceil(numpy.log2(n)))

def lagrange_invariant(gausslet):
    base_ray = gausslet.base_ray
    origin = numpy.asarray(base_ray.origin)
    direction = numpy.asarray(base_ray.direction)
    axis1 = numpy.asarray(base_ray.E_vector)
    axis2 = numpy.cross(axis1, direction)
    
    paras = gausslet.parabasal_rays
    
    def get_origin(porigin):
        return numpy.asarray(porigin) - origin
    
    def get_direction(pdir):
        return numpy.asarray(pdir)# - direction
    
    h1 = [(get_origin(p.origin)*axis1).sum() for p in paras]
    h2 = [(get_origin(p.origin)*axis2).sum() for p in paras]
    u1 = [(get_direction(p.direction)*axis1).sum() for p in paras]
    u2 = [(get_direction(p.direction)*axis2).sum() for p in paras]
    
    def HU(i,j):
        return (h1[i]*u1[j] + h2[i]*u2[j]) - ( h1[j]*u1[i] + h2[j]*u2[i] )
    
    #out = [HU(0,3), HU(1,2), HU(0,5), HU(1,4), HU(3,4), HU(2,5)]
    out = [HU(0,5), HU(1,2), HU(3,4), HU(1,4), HU(0,3), HU(2,5)]
#     out = numpy.zeros((6,6),'d')
#     for i in range(6):
#         for j in range(6):
#             out[i,j] = 
            
    return out


def make_hexagonal_grid(radius, spacing=1.0, connectivity=False):
    """Creates a 2d hexagonal grid.
    Radius and spacing have the same units.
    Returns - x,y: 1d arrays of the respective coordinate.
    if connectivity=True:
        Returns - x, y, neighbours: where neighbours is an (N,6) array of indices giving the adjacent neighbours to each coordinate
    """
    
    cosV = numpy.cos(numpy.pi*30/180.)
    sinV = numpy.sin(numpy.pi*30/180.)
    
    nsteps = int(radius/(spacing*numpy.cos(numpy.pi*30/180.)))
    i = numpy.arange(-nsteps-1, nsteps+2)
    j = numpy.arange(-nsteps-1, nsteps+2)
    
    vi, vj = numpy.meshgrid(i,j)
    
    ni, nj = vi.shape
    vi.shape = -1
    vj.shape = -1
    
    x = (vi + sinV*vj)*spacing
    y = cosV*vj*spacing
    
    r2 = x**2 + y**2
    select = r2 < (radius**2)
    
    if connectivity:
        labels = -numpy.ones((ni+2, nj+2), numpy.int)
        labels[1:-1, 1:-1] = numpy.arange(ni*nj).reshape(ni,nj)
        
        centres = -numpy.ones((vi.size,), numpy.int)
        neighbours = numpy.dstack([labels[2:,1:-1], labels[:-2,1:-1],
                                   labels[1:-1,2:], labels[1:-1,:-2],
                                   labels[2:,:-2], labels[:-2,2:]])
        neighbours.shape = (-1,6)
        
        sel = centres
        centres[select] = numpy.arange(select.sum())
        nb = centres[numpy.compress(select, neighbours, axis=0)]
        
        return x[select], y[select], nb
    else:
        return x[select], y[select]


def decompose_angle(origin, direction, axis1, E_field, input_spacing, max_angle, wavelength, 
                    oversample=4, E_max=None, pos_max=None):
    """
    Compute a set of Gausslets over a range of directions, where the gausslet amplitude is
    obtained by FFT of the input E-field profile. The Gausslets all have an origin at the given
    origin point.
    
    params:
        origin - a (x,y,z) position vector for the origin of the Gausslets and the centre of the E-field
                distribution
        direction - a direction vector giving the output optical axis
        axis1 - a vector orthogonal to the direction giving the E1 polarisation axis and the direction of the E-field
                array 1st axis.
        E_field - a complex array of shape (N,M,3). The vector E-field.
        input_spacing - a scalar giving the sample spacing for the input E-field, in microns
        max_angle - sets the maximum output angle for the outgoing gausslets. Rays outside this angle are omitted. Units=degrees
        wavelength - sets the wavelength for the source, in microns
        oversample - sets the amount of padding added to the E_field data to increase the oversampling of the output.
    """
    input_spacing /= 1000.0
    wavelength /= 1000.0
    direction = normaliseVector(direction)
    d2 = normaliseVector(numpy.cross(direction, axis1))
    d1 = numpy.cross(direction, d2)
    
    N,M = E_field.shape[:2]
    target_size = next_power_of_two(max(N,M)) * oversample
    
    n_start = int((target_size/2) - (N/2))
    m_start = int((target_size/2) - (M/2))
    
    data_in = numpy.zeros((target_size,target_size,3), dtype=numpy.complex128)
    
    data_in[n_start:n_start+N, m_start:m_start+M, :] = E_field
    data_in = fft.ifftshift(data_in, axes=(0,1))
    
    data_out = fft.fft2(data_in, axes=(0,1)) / (target_size**2)
    data_out = fft.fftshift(data_out, axes=(0,1))
        
    ###The individual gausslet beam-waist radii at the origin are determined from the
    ### angular spacing of the output gausslets, and the wavelength.
    kmax = 2.0*numpy.pi/(2*input_spacing) #
    k_abs = 2.0*numpy.pi/wavelength #where wavelength is in microns
    
    kr = numpy.linspace(-kmax,kmax,data_out.shape[0]+1)[:-1]
    
    k_limit = k_abs*numpy.sin(numpy.pi*max_angle/180.0)
    k_grid_spacing = (kr[1]-kr[0])*oversample #The spacing for the hexagonal grid
    
    kx,ky = make_hexagonal_grid(k_limit, spacing=k_grid_spacing)
    _x = kx
    _y = ky
    
    kz = numpy.sqrt(k_abs**2 - kx**2 - ky**2)
        
    directions = (kx[:,None]*d1 + ky[:,None]*d2 + kz[:,None]*direction)/k_abs
    #directions = normaliseVector(directions)
    #print("interp:", x.min(),x.max(),y.min(),y.max(), kmax, kz, k_limit)
    
    Ex_real = RectBivariateSpline(kr,kr,data_out[:,:,0].real)(_x, _y, grid=False)
    Ex_imag = RectBivariateSpline(kr,kr,data_out[:,:,0].imag)(_x, _y, grid=False)
    Ey_real = RectBivariateSpline(kr,kr,data_out[:,:,1].real)(_x, _y, grid=False)
    Ey_imag = RectBivariateSpline(kr,kr,data_out[:,:,1].imag)(_x, _y, grid=False)
    Ez_real = RectBivariateSpline(kr,kr,data_out[:,:,2].real)(_x, _y, grid=False)
    Ez_imag = RectBivariateSpline(kr,kr,data_out[:,:,2].imag)(_x, _y, grid=False)
    
    E_real = numpy.column_stack((Ex_real, Ey_real, Ez_real))
    E_imag = numpy.column_stack((Ex_imag, Ey_imag, Ez_imag))
    
    E_full = -(-1.0j*E_real + E_imag)
    
    E_vectors = normaliseVector(numpy.cross(directions, d2))
    E2_vectors = normaliseVector(numpy.cross(directions, E_vectors))
    
    E1_amp = (E_vectors * E_full).sum(axis=1)
    E2_amp = (E2_vectors * E_full).sum(axis=1)
    
    ray_data = numpy.zeros(directions.shape[0], dtype=ray_dtype)
    
    gausslet_radius = wavelength*k_abs/(k_grid_spacing*numpy.pi)
    
    ### Some sort of guess to get the output power roughly right.
    scaling = numpy.sqrt(2)*input_spacing*target_size
    
    ray_data['origin'] = origin + ((d1+d2)*(input_spacing*0.5))
    ray_data['direction'] = directions
    ray_data['wavelength_idx'] = 0
    ray_data['E_vector'] = E_vectors
    ray_data['E1_amp'] = E1_amp*scaling
    ray_data['E2_amp'] = E2_amp*scaling
    ray_data['refractive_index'] = 1.0+0.0j
    ray_data['normal'] = [[0,1,0]]
    ray_data['phase'] = 0.0
    rays = GaussletCollection.from_rays(ray_data)
    rays.wavelengths = wl = numpy.array([wavelength])
    working_dist=0.0
    
    print("Gausslet radius:", gausslet_radius)
    rays.config_parabasal_rays(wl, gausslet_radius, working_dist)
    rays.wavelengths = wl
    
    if E_max is not None and pos_max is not None:
        print("E_max:", E_max, "pos_max:", pos_max)
        e_test = eval_Efield_from_gausslets(rays, pos_max[None,:], blending=1.0)
        print("e_test:", e_test)
        scaling = (E_max/e_test).mean()
        #ray_data['E1_amp'] *= scaling
        #ray_data['E2_amp'] *= scaling
        
        rays = GaussletCollection.from_rays(ray_data)
        rays.wavelengths = wl
        rays.config_parabasal_rays(wl, gausslet_radius, working_dist)
    
    return rays, data_out


def decompose_position(input_rays, origin, direction, axis1, radius, resolution, curvature=None, blending=1.5):
    """
    Spatially decompose the input Gausslets into a new set of Gausslets defined over a circular area of the 
    given radius. The output ray density is given by the resolution parameter.
    
    Parameters
    ----------
    
    input_rays : GaussletCollection
                 The input rays for the decomposition
                 
    origin : tuple, (float, float, float)
             The centre of the decomposition plane
             
    direction : tuple (float, float, float)
                The normal-vector for the decomposition plane
                
    axis1 : tuple (float, float, float)
            A vector defining the orientation of the decomposition plane (orthogonal to direction)
                
    radius : float
             The maximum radius giving the limit of the output rays
             
    resolution : float
                 The number of rays generated between the decomposition plane centre and radius edge
                 
    curvature : float, optional
                The approximate radius-of-curvature of the wavefront. I.e. distance to focus. None=Inf.
                
    Returns
    -------
    
    rays : GaussletCollection
            The out-going Gausslets
    """
    spacing = radius / resolution
    wavelengths = input_rays.wavelengths
    
    if len(numpy.unique(wavelengths)) > 1:
        raise ValueError("Can't decompose wavefront with more than one wavelength present.")
    
    wavelength = wavelengths[0]
    
    origin = numpy.asarray(origin)
    direction = normaliseVector(numpy.asarray(direction))
    axis1 = normaliseVector(numpy.asarray(axis1))
    axis2 = normaliseVector(numpy.cross(axis1, direction))
    axis1 = numpy.cross(axis2, direction)
    
    ###In this version, we will evaluate the field on a cartesian grid
    _radius = radius * (1 + 2./resolution)
    x_ = y_ = numpy.linspace(-_radius,_radius, int((1.1*resolution)*2))
    
    x,y = numpy.meshgrid(x_, y_)
    
    origins_in = origin[None,:] + x.reshape(-1,)[:,None]*axis1[None,:] + y.reshape(-1)[:,None]*axis2[None,:]
    
    efield = EFieldSummation(input_rays)
    E_field = efield.evaluate(origins_in)
    
    ### Project onto local axes to get orthogonal polarisations and choose the one with the most power
    E1_amp = (E_field*axis1[None,:]).sum(axis=1)
    E2_amp = (E_field*axis2[None,:]).sum(axis=1)
    
    P1 = (E1_amp.real**2 + E1_amp.imag**2).sum()
    P2 = (E2_amp.real**2 + E2_amp.imag**2).sum()
    
    if P1 > P2:
        ### Always in the range -pi to +pi
        phase = numpy.arctan2(E1_amp.imag, E1_amp.real)
    else:
        phase = numpy.arctan2(E2_amp.imag, E2_amp.real)
    
    nx = len(x_)
    ny = len(y_)
    phase.shape = (nx, ny)
    
    if curvature is not None:
        sphz = -(curvature - numpy.sqrt(curvature*curvature - x*x - y*y))
        sph = sphz*2000.0*numpy.pi/wavelength - numpy.pi
        phase = phase - sph
        phase = phase%(2*numpy.pi)
        phase = phase - numpy.pi
    else:
        sph = 0
    
    uphase, residuals = unwrap2d(phase, anchor=(nx//2, ny//2))
    uphase2 = uphase + sph
    
    ### Setting s=0 results in oscillatory behaviour near the edge along the x-idrection.
    wavefront = RectBivariateSpline(x_, y_, -uphase2*wavelength/(2000*numpy.pi), kx=3, ky=3, s=0.001)
    
    ### Now make a hex-grid of new ray start-points
    rx,ry, nb = make_hexagonal_grid(radius, spacing=spacing, connectivity=True)
    
    N = len(rx)
    
    origins = origin[None,:] + rx[:,None]*axis1[None,:] + ry[:,None]*axis2[None,:]
    
    ### First derivatives give us the ray directions
    dx = wavefront(rx,ry,dx=1, grid=False)
    dy = wavefront(rx,ry,dy=1, grid=False)
    
    ### Get second derivatives, to get wavefront curvature
    dx2 = wavefront(rx,ry,dx=2, grid=False)
    dy2 = wavefront(rx,ry,dy=2, grid=False)
    dxdy = wavefront(rx,ry,dx=1,dy=1, grid=False)

    ###Convert to A,B,C coefficients
    A,B,C,xl,yl,zl = calc_mode_curvature(rx, ry, dx, dy, dx2, dy2, dxdy)
    
    E_in = efield.evaluate(origins) #should be a (N,3) array of complex values
    
    directions = axis1[None,:]*zl[:,0,None] + axis2[None,:]*zl[:,1,None] + direction[None,:]*zl[:,2,None]
    E_vectors = axis1[None,:]*xl[:,0,None] + axis2[None,:]*xl[:,1,None] + direction[None,:]*xl[:,2,None]
    H_vectors = axis1[None,:]*yl[:,0,None] + axis2[None,:]*yl[:,1,None] + direction[None,:]*yl[:,2,None]
        
    E1_amp = (E_in*E_vectors).sum(axis=-1)
    E2_amp = (E_in*H_vectors).sum(axis=-1)
    
    ray_data = numpy.zeros(N, dtype=ray_dtype)
    
    ray_data['origin'] = origins
    ray_data['direction'] = directions
    ray_data['E_vector'] = E_vectors
    ray_data['refractive_index'] = 1.0
    ray_data['E1_amp'] = E1_amp
    ray_data['E2_amp'] = E2_amp
    
    gausslets = GaussletCollection.from_rays(ray_data)
    gausslets.wavelengths = wavelengths
    gausslets.config_parabasal_rays(wavelengths, spacing/blending, 0.0)
    apply_mode_curvature(gausslets, -A, -B, -C)
    
    E_test = eval_Efield_from_gausslets(gausslets, origins)
    
    power_scaling = (E_in.real**2 + E_in.imag**2).sum() / (E_test.real**2 + E_test.imag**2).sum()
    
    gausslets.scale_amplitude(numpy.sqrt(power_scaling))
    
    return gausslets, E_field, uphase
    


"""
Utility module
"""

from numpy.linalg import solve
import numpy as np
from raypier.core.ctracer import RayCollection, GaussletCollection


def find_ray_focus(ray_collection):
    if isinstance(ray_collection, GaussletCollection):
        data = ray_collection.copy_as_array()
        ray = data['base_ray']
        E1 = ray['E1_amp']
        E2 = ray['E2_amp']
        origin = ray['origin']
        direction = ray['direction']
    else:
        if not isinstance(ray_collection, RayCollection):
            raise TypeError("Expecting a RayCollection object")
        E1 = ray_collection.E1_amp
        E2 = ray_collection.E2_amp
        origin = ray_collection.origin
        direction = ray_collection.direction
    weights = E1.real**2 + E1.imag**2 + E2.real**2 + E2.imag**2
    weights /= weights.sum()
    return find_focus(origin, direction, weights=weights)


def find_focus(ray_origins, ray_directions, weights=None):
    """
    givens Nx3 arrays for ray origin and directions, return the 
    3D point of closest intersection, found in the least-squares 
    sense.
    
    See https://math.stackexchange.com/q/1762491
    """
    
    o = ray_origins
    n = ray_directions / (ray_directions**2).sum(axis=1, keepdims=True)
    
    A = np.matmul(n[:,:,np.newaxis], n[:,np.newaxis,:]) - np.eye(3)[np.newaxis,:,:]
    # A.shape = (N,3,3), for N rays
    if weights is not None:
        A *= weights[:,None,None]
        
    AA = A.sum(axis=0) #AA is a (3,3) shape matrix
    
    b = np.matmul(A,o[:,:,np.newaxis]).sum(axis=0)
    
    out = solve(AA,b[:,0])
    return out


if __name__=="__main__":
    directions = np.array([[ 0.00000000e+00,  1.00000000e+00,  3.58332104e-17],
       [ 2.00168533e-01,  9.79761480e-01,  3.04549358e-17],
       [ 1.22567877e-17,  9.79761480e-01,  2.00168533e-01],
       [-2.00168533e-01,  9.79761480e-01,  5.49685112e-17],
       [-3.67703631e-17,  9.79761480e-01, -2.00168533e-01]])
    origins = np.array([[ 0.00000000e+00, -1.40000000e+01,  1.02258724e-15],
       [-9.83472949e+00, -1.51142336e+01,  8.82453980e-16],
       [-6.02203500e-16, -1.51142336e+01, -9.83472949e+00],
       [ 9.83472949e+00, -1.51142336e+01, -3.21953019e-16],
       [ 1.80661050e-15, -1.51142336e+01,  9.83472949e+00]])
    
    print(directions.shape, origins.shape)
    out = find_focus(origins, directions)
    print(out)
    
    ###Now check numerically
    def get_errs(y_coord):
        y = np.array([[0.0,1.0,0.0]])
        y_ = y_coord - origins[:,1] 
        det = np.dot(directions, y[0])
        a = y_/det
        intersection = origins + (a.reshape(-1,1)*directions)
        centroid = intersection.mean(axis=0)
        delta = ((intersection - centroid)**2).sum()
        return delta
    
    #for i in np.linspace(20,40,100):
    #    print i, get_errs(i)
        
    from matplotlib import pyplot as pp
    
    a=np.linspace(0.0,100.0,200)
    
    for i in range(5):
        x = origins[i,0] + a * directions[i,0]
        y = origins[i,1] + a * directions[i,1]
    
        pp.plot(y,x)
    pp.show()
    
    
